__author__ = 'giacomov'

import collections
import os
import warnings

from astromodels.sources.source import POINT_SOURCE, EXTENDED_SOURCE, PARTICLE_SOURCE

from astromodels.my_yaml import my_yaml
from astromodels.utils.disk_usage import disk_usage
from astromodels.utils.table import dict_to_table
from astromodels.parameter import Parameter, IndependentVariable
from astromodels.tree import Node
from astromodels.functions.function import get_function


class ModelFileExists(IOError):
    pass


class InvalidInput(ValueError):
    pass


class CannotWriteModel(IOError):
    def __init__(self, directory, message):
        # Add a report on disk usage to the message

        free_space = disk_usage(directory).free

        message += "\nFree space on the file system hosting %s was %.2f Mbytes" % (
            directory, free_space / 1024.0 / 1024.0)

        super(CannotWriteModel, self).__init__(message)


class Model(Node):

    def __init__(self, *sources):

        # Setup the node

        super(Model, self).__init__("root")

        # Dictionary to keep point sources

        self._point_sources = collections.OrderedDict()

        # Dictionary to keep extended sources

        self._extended_sources = collections.OrderedDict()

        # Dictionary to keep particle sources

        self._particle_sources = collections.OrderedDict()

        # Loop over the provided sources and process them

        for source in sources:

            self._add_child(source)

            # Now see if this is a point or extended source, and add them to the
            # appropriate dictionary

            if source.source_type == POINT_SOURCE:

                self._point_sources[source.name] = source

            elif source.source_type == EXTENDED_SOURCE:

                self._extended_sources[source.name] = source

            elif source.source_type == PARTICLE_SOURCE:

                self._particle_sources[source.name] = source

            else:

                raise InvalidInput("Input sources must be either a point source or an extended source")

        # Also create the lists for the point and extended sources, for speed

        self._point_sources_list = self._point_sources.values()
        self._extended_sources_list = self._extended_sources.values()
        self._particle_sources_list = self._particle_sources.values()

        # Now make the list of all the existing parameters

        self._update_parameters()

    def _update_parameters(self):

        self._parameters = self._find_instances(Parameter)

    @property
    def parameters(self):
        """
        Return a dictionary with all parameters

        :return: dictionary of parameters
        """
        self._update_parameters()

        return self._parameters

    @property
    def free_parameters(self):
        """
        Get a dictionary with all the free parameters in this model

        :return: dictionary of free parameters
        """

        # Refresh the list

        self._update_parameters()

        # Filter selecting only free parameters

        free_parameters_dictionary = collections.OrderedDict()

        for parameter_name, parameter in self._parameters.iteritems():

            if parameter.free:

                free_parameters_dictionary[parameter_name] = parameter

        return free_parameters_dictionary

    @property
    def linked_parameters(self):
        """
        Get a dictionary with all parameters in this model in a linked status. A parameter is in a linked status
        if it is linked to another parameter (i.e. it is forced to have the same value of the other parameter), or
        if it is linked with another parameter or an independent variable through a law.

        :return: dictionary of linked parameters
        """

        # Refresh the list

        self._update_parameters()

        # Filter selecting only free parameters

        linked_parameter_dictionary = collections.OrderedDict()

        for parameter_name, parameter in self._parameters.iteritems():

            if parameter.has_auxiliary_variable():

                linked_parameter_dictionary[parameter_name] = parameter

        return linked_parameter_dictionary

    def __getitem__(self, path):
        """
        Get a parameter from a path like "source_1.component.powerlaw.logK". This might be useful in certain
        context, although in an interactive analysis there is no reason to use this.

        :param path: the address of the parameter
        :return: the parameter
        """

        return self._get_child_from_path(path)

    def __contains__(self, path):
        """
        This allows the model to be used with the "in" operator, like;

        > if 'myparameter' in model:
        >    print("Myparameter is contained in the model")

        :param path: the parameter to look for
        :return:
        """

        try:

            _ = self._get_child_from_path(path)

        except (AttributeError, KeyError):

            return False

        else:

            return True

    @property
    def point_sources(self):
        return self._point_sources

    @property
    def extended_sources(self):
        return self._extended_sources

    @property
    def particle_sources(self):
        return self._particle_sources

    def add_independent_variable(self, variable):
        """
        Add a global independent variable to this model, such as time.

        :param variable: an IndependentVariable instance
        :return: none
        """

        assert isinstance(variable, IndependentVariable),"Variable must be an instance of IndependentVariable"

        if variable.name in self._children:

            self._remove_child(variable.name)

        self._add_child(variable)

    def link(self, parameter_1, parameter_2, link_function=None):
        """
        Link the value of the provided parameters through the provided function (identity is the default, i.e.,
        parameter_1 = parameter_2).

        :param parameter_1: the first parameter
        :param parameter_2: the second parameter
        :param link_function: a function instance. If not provided, the identity function will be used by default.
        Otherwise, this link will be set: parameter_1 = link_function(parameter_2)
        :return: (none)
        """

        if link_function is None:
            # Use the identity function by default

            link_function = get_function('identity')

        parameter_1.add_auxiliary_variable(parameter_2, link_function)

        # Now set the units of the link function
        link_function.set_units(parameter_2.unit, parameter_1.unit)

    def unlink(self, parameter):
        """
        Sets free a parameter which has been linked previously

        :param parameter: the parameter to be set free
        :return: (none)
        """

        if parameter.has_auxiliary_variable():

            parameter.remove_auxiliary_variable()

        else:

            warnings.warn("Parameter %s has no link to be removed." % parameter.path)

    def _repr__base(self, rich_output=False):

        if rich_output:

            div = '<br>'

        else:

            div = '\n'

        # Print the name of point sources if there are any

        representation = "Point sources: "

        if self._point_sources:

            representation += ",".join(self._point_sources.keys())

        else:

            representation += "(none)"

        # Print the name of extended sources if there are any

        representation += "%s%sExtended sources: " % (div, div)

        if self._extended_sources:

            representation += ",".join(self._extended_sources.keys())

        else:

            representation += "(none)"

        # Print the name of extended sources if there are any

        representation += "%s%sParticle sources: " % (div, div)

        if self._particle_sources:

            representation += ",".join(self._particle_sources.keys())

        else:

            representation += "(none)"

        representation += "%s%sFree parameters:%s" % (div, div, div)

        parameters_dict = collections.OrderedDict()

        for parameter_name, parameter in self.free_parameters.iteritems():

            # Generate table with only a minimal set of info

            parameters_dict[parameter_name] = parameter.to_dict()

        table = dict_to_table(parameters_dict, ['value','unit','min_value','max_value','delta','free'])

        if rich_output:

            representation += table._repr_html_()

        else:

            representation += table.__repr__()

            representation += div

        # Print linked parameters

        linked_parameters = self.linked_parameters

        if linked_parameters:

            representation += "%s%sLinked parameters:%s" % (div, div, div)

            parameters_dict = collections.OrderedDict()

            for parameter_name, parameter in linked_parameters.iteritems():

                # Generate table with only a minimal set of info

                variable, law = parameter.auxiliary_variable

                this_dict = collections.OrderedDict()

                this_dict['linked to'] = variable.path
                this_dict['function'] = law.name
                this_dict['current value'] = parameter.value.value
                this_dict['unit'] = parameter.unit

                parameters_dict[parameter_name] = this_dict

            table = dict_to_table(parameters_dict)

            if rich_output:

                representation += table._repr_html_()

            else:

                representation += table.__repr__()

                representation += div

        return representation

    def save(self, output_file, overwrite=False):

        """Save the model to disk"""

        if os.path.exists(output_file) and overwrite is False:

            raise ModelFileExists("The file %s exists already. If you want to overwrite it, use the 'overwrite=True' "
                                  "options as 'model.save(\"%s\", overwrite=True)'. " % (output_file, output_file))

        else:

            # Get the serialization dictionary

            data = self.to_dict()

            # Add the types to the sources

            for key in data.keys():

                try:

                    element = self._children[key]

                except KeyError:

                    raise RuntimeError("Source %s is unknown" % key)

                else:

                    # There are two possible cases. Either the element is a source, or it is an independent
                    # variable

                    if hasattr(element, 'source_type'):

                        # Change the name of the key adding the source type

                        data['%s (%s)' % (key, element.source_type)] = data.pop(key)

                    elif isinstance(element, IndependentVariable):

                        data['%s (%s)' % (key, 'IndependentVariable')] = data.pop(key)

                    else:

                        raise CannotWriteModel("Found an unknown class at the top level")

            # Write it to disk

            try:

                # Get the YAML representation of the data

                representation = my_yaml.dump(data)

                with open(output_file, "w+") as f:

                    # Add a new line at the end of each voice (just for clarity)

                    f.write(representation.replace("\n", "\n\n"))

            except IOError:

                raise CannotWriteModel(os.path.dirname(os.path.abspath(output_file)),
                                       "Could not write model file %s. Check your permissions to write or the "
                                       "report on the free space which follows: " % output_file)

    def get_number_of_point_sources(self):
        """
        Return the number of point sources

        :return: number of point sources
        """
        return len(self._point_sources)

    def get_point_source_position(self, id):
        """
        Get the point source position (R.A., Dec)

        :param id: id of the source
        :return: a tuple with R.A. and Dec.
        """

        pts = self._point_sources_list[id]

        return pts.position.get_ra(), pts.position.get_dec()

    def get_point_source_fluxes(self, id, energies):
        """
        Get the fluxes from the id-th point source

        :param id: id of the source
        :param energies: energies at which you need the flux
        :return: fluxes
        """

        return self._point_sources_list[id](energies)

    def get_point_source_name(self, id):

        return self._point_sources_list[id].name

    def get_number_of_extended_sources(self):
        """
        Return the number of extended sources

        :return: number of extended sources
        """
        return len(self._extended_sources)

    def get_extended_source_fluxes(self, id, j2000_ra, j2000_dec, energies):
        """
        Get the flux of the id-th extended sources at the given position at the given energies

        :param id: id of the source
        :param j2000_ra: R.A. where the flux is desired
        :param j2000_dec: Dec. where the flux is desired
        :param energies: energies at which the flux is desired
        :return: flux array
        """

        return self._extended_sources_list[id](j2000_ra, j2000_dec, energies)

    def get_extended_source_name(self, id):
        """
        Return the name of the n-th extended source

        :param id: id of the source (integer)
        :return: the name of the id-th source
        """

        return self._extended_sources_list[id]

    def get_extended_source_boundaries(self, id):

        (ra_min, ra_max), (dec_min, dec_max) = self._extended_sources_list[id].get_boundaries()

        return ra_min, ra_max, dec_min, dec_max

    def is_inside_any_extended_source(self, j2000_ra, j2000_dec):

        return True

    def get_number_of_particle_sources(self):
        """
        Return the number of particle sources

        :return: number of particle sources
        """

        return len(self._particle_sources)

    def get_particle_source_fluxes(self, id, energies):
        """
        Get the fluxes from the id-th point source

        :param id: id of the source
        :param energies: energies at which you need the flux
        :return: fluxes
        """

        return self._particle_sources_list[id]._get_flux(energies)

    def get_particle_source_name(self, id):

        return self._particle_sources_list[id].name


import numpy as np


def angular_distance(ra1, dec1, ra2, dec2):
    """
    Returns the angular distance between two points, two sets of points, or a set of points and one point.

    :param ra1: array or float, longitude of first point(s)
    :param dec1: array or float, latitude of first point(s)
    :param ra2: array or float, longitude of second point(s)
    :param dec2: array or float, latitude of second point(s)
    :return: angular distance(s) in degrees
    """

    # Vincenty formula, slower than the Haversine formula in some cases, but stable also at antipodes

    lon1 = np.deg2rad(ra1)
    lat1 = np.deg2rad(dec1)
    lon2 = np.deg2rad(ra2)
    lat2 = np.deg2rad(dec2)

    sdlon = np.sin(lon2 - lon1)
    cdlon = np.cos(lon2 - lon1)
    slat1 = np.sin(lat1)
    slat2 = np.sin(lat2)
    clat1 = np.cos(lat1)
    clat2 = np.cos(lat2)

    num1 = clat2 * sdlon
    num2 = clat1 * slat2 - slat1 * clat2 * cdlon
    denominator = slat1 * slat2 + clat1 * clat2 * cdlon

    return np.rad2deg(np.arctan2(np.sqrt(num1 ** 2 + num2 ** 2), denominator))



__author__ = 'giacomov'

""" Implements a dictionary for which assignment can be lock, so that an operation like 'd[key] = value' will fail.
The main reason for this is the 'parameters' dictionary for the models. To assign a float 'value' to a parameter
the syntax to be used is parameters[parname].value = value and not parameters[parname] = value, which would overwrite
the parameter class with the float. """

import collections


class ParameterDictionary(collections.OrderedDict):

    __is_readonly = False

    def lock(self):

        self.__is_readonly = True

    def unlock(self):

        self.__is_readonly = False

    def __setitem__(self, *args, **kwargs):

        if not self.__is_readonly:

            super(ParameterDictionary, self).__setitem__(*args, **kwargs)

        else:

            raise ValueError("To assign values to the parameters use 'parameters[parname].value = value', and not "
                             "'parameters[parname] = value'")






# Import the version
from version import __version__

from .sources.point_source import PointSource
from .sources.extended_source import ExtendedSource
from .sources.particle_source import ParticleSource
from parameter import Parameter, IndependentVariable, SettingOutOfBounds
from .functions.functions import *
from .functions.functions_2D import *
from .functions.functions_3D import *
from .functions.function import list_functions, get_function_class
from model import Model
from spectral_component import SpectralComponent
from model_parser import load_model
from units import get_units
from .xspec.factory import setup_xspec_models, has_xspec
import astropy.units as u

astromodels_units = get_units()

if has_xspec:

    from .xspec.factory import *
    from .xspec.xspec_settings import *


__author__ = 'giacomov'

import ast

import numpy


class ExpressionCannotBeParsed(Exception):
    pass


class ErrorInFormula(Exception):
    pass


class Formula(object):
    def __init__(self, formula):
        """

        :param formula: the formula to be parsed. Must be a string with a simple one-line expression. You can use
        anything supported natively by python (like operations as +,-,*,/) and any mathematical function from numpy.
        For example: 'k * exp(x)' express an exponential function with only one parameter, i.e., the normalization k;
        'k * power(x,index)' or equivalently 'k* x**index' express a power law with index 'index' and normalization k.
        A list of mathematical function contained in numpy can be found in the numpy documentation:
        http://docs.scipy.org/doc/numpy/reference/routines.math.html .

        Some rules:
        * x,y,z are variables, anything else is assumed to be either a parameter or a function. So for example
        'k * exp(x)' has one variable (x) and one parameter (k). 'k * exp(a * x + y)' is a function of two variables
        (x,y) and has two parameters (k and a).
        * the expression must be a valid python expression
        * must return a result of the same size of the input
        * always assume you are dealing with np arrays

        :return:
        """

        self._formula = formula

        try:

            self._tree = ast.parse(formula)

        except SyntaxError:

            raise ExpressionCannotBeParsed("The expression '%s' could not be parsed" % formula)

        # Traverse the tree and get the name of the functions which are used in the formula

        self._functions = []

        for node in ast.walk(self._tree):

            if isinstance(node, ast.Call):

                # This is a function call

                # Check that it is not an attribute of the type np.exp

                if isinstance(node.func, ast.Attribute):

                    # Check if the user has expressed a function as np.exp instead of exp

                    if isinstance(node.func.value, ast.Name):

                        if node.func.value.id == 'np' or node.func.value.id == 'numpy':

                            # Advice to use only the function instead of np.function

                            v = "%s.%s" % (node.func.value.id, node.func.attr)

                            raise ErrorInFormula("Do not express numpy math function as '%s'. "
                                                 "Use just '%s' instead." % (v, node.func.attr))

                        else:

                            raise ErrorInFormula("You cannot use modules such as '%s' in a formula"
                                                 % node.func.value.id)

                    # Check if this is something like scipy.stats.norm.pdf (i.e., multiple attributes)

                    if isinstance(node.func.value, ast.Attribute):
                        raise ErrorInFormula("In a formula you can only use math functions from numpy")

                # Now check if this function is really a numpy function

                if not hasattr(numpy, node.func.id):

                    raise ErrorInFormula("The function %s is not a valid numpy math function" % node.func_id)

                else:

                    # Check that the function is a ufunc

                    if not isinstance(getattr(numpy, node.func.id), numpy.ufunc):
                        raise ErrorInFormula("The function %s is not a numpy ufunc" % node.func_id)

                self._functions.append(node.func.id)

        # Traverse the tree and get variables and parameters

        self._names = []

        for node in ast.walk(self._tree):

            if isinstance(node, ast.Name):

                # Since functions are still Name instances, check first if we didn't save this already as a function

                if node.id in self._functions:

                    continue

                else:

                    # This is a variable. Add it to the list if it is not there already

                    if node.id not in self._names:
                        self._names.append(node.id)

        # Now divide parameters from variables such as x,y,z

        self._variables = filter(lambda token: token in ['x', 'y', 'z'], self._names)

        # Safety check: the user must name the variables in order. In other words, if there is a y variable there
        # MUST be a x as well. If there is a z variable, there must be both x and y as well.

        if 'z' in self._variables and ('x' not in self._variables or 'y' not in self._variables):
            raise ErrorInFormula("The first variable must be called 'x', the second 'y' and the third 'z'."
                                 " If you have 'z' you must have also 'x' and 'y'. Rename your variables accordingly.")

        if 'y' in self._variables and 'x' not in self._variables:
            raise ErrorInFormula("You have 'y' in your formula but you don't have 'x'. Rename your variable as 'x'.")

        self._parameters = filter(lambda token: token not in ['x', 'y', 'z'], self._names)

    @property
    def variables(self):
        """

        :return: list of the variables such as x,y,z
        """
        return self._variables

    @property
    def dimensionality(self):
        """

        :return: return the number of variables the expression is a function of.
        """
        return len(self.variables)

    @property
    def parameters(self):
        """

        :return: list of parameters
        """
        return self._parameters

    @property
    def number_of_parameters(self):
        """

        :return: number of parameters
        """
        return len(self.parameters)

    @property
    def functions(self):
        """

        :return: list of functions used in the formula
        """
        return self._functions

    @property
    def formula(self):
        """

        :return: the formula as a string
        """

        return self._formula

    def __repr__(self):

        msg = "Function of %s variables with parameters %s:\n%s" % (self.dimensionality,
                                                                    ",".join(self.parameters),
                                                                    self._formula)

        return msg



import astropy.units as u

from astromodels.functions.function import Function3D, FunctionMeta

import numpy as np

from astromodels.utils.angular_distance import angular_distance


class Continuous_injection_diffusion(Function3D):
    r"""
        description :

            Positron and electrons diffusing away from the accelerator

        latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            rdiff0 :

                desc : Projected diffusion radius. The maximum allowed value is used to define the truncation radius.
                initial value : 1.0
                min : 0
                max : 20

            delta :

                desc : index for the diffusion coefficient
                initial value : 0.5
                min : 0.3
                max : 0.6
                fix : yes

            piv :

                desc : Pivot
                initial value : 2e10
                min : 0
                fix : yes

            piv2 :
                desc : Pivot for the diffusion radius
                initial value : 1e9
                min : 0
                fix : yes

        """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, delta, piv, piv2):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.
        try:

            rdiff = rdiff0 * np.power(energy / piv, (delta - 1.) / 2. * (0.54 + 0.046 * np.log10(energy / piv2)))

        except ValueError:

            # This happens when using units, because astropy.units fails with the message:
            # "ValueError: Quantities and Units may only be raised to a scalar power"

            # Work around the problem with this loop, which is slow but using units is only for testing purposes or
            # single calls, so it shouldn't matter too much
            rdiff = np.array( map(lambda x: (rdiff0 * np.power(energy/ piv, x)).value,
                                  (delta - 1.) / 2. * (0.54 + 0.046 * np.log10(energy / piv2)))) * rdiff0.unit

        angsep = angular_distance(lon, lat, lon0, lat0)

        pi = np.pi

        return np.power(180.0 / pi, 2) * 1.2154 / (pi * np.sqrt(pi) * rdiff * (angsep + 0.06 * rdiff)) * \
               np.exp(-np.power(angsep, 2) / rdiff ** 2)


    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90., self.lat0.value - maximum_rdiff)
        max_latitude = min(90., self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if max_abs_lat > 89. or maximum_rdiff / np.cos(max_abs_lat * np.pi / 180.) >= 180.:

            min_longitude = 0.
            max_longitude = 360.

        else:

            min_longitude = self.lon0.value - maximum_rdiff / np.cos(max_abs_lat * np.pi / 180.)
            max_longitude = self.lon0.value + maximum_rdiff / np.cos(max_abs_lat * np.pi / 180.)

            if min_longitude < 0.:

                min_longitude += 360.

            elif max_longitude > 360.:

                max_longitude -= 360.

        return (min_longitude, max_longitude), (min_latitude, max_latitude), (0, np.inf)


import collections
import os
import re
import sys
import warnings
import astropy.units as u

from astromodels.my_yaml import my_yaml
from astromodels.functions.function import FunctionMeta, Function1D, get_function_class

try:

    from astromodels.xspec import _xspec

except ImportError:

    has_xspec = False

else:

    has_xspec = True


class XSpecNotAvailable(UserWarning):
    pass


# This list defines all python protected names (names which variables or attributes should not have)
illegal_variable_names_ = 'and, assert, break, class, continue, def, del, elif, else, except, exec, finally, for,' \
                         'if, import, in, is, lambda, not, or, pass, print, raise, return, try, while, data, ' \
                         'float, int, numeric, array, close, input, open, range, type, write, zeros, from, global'
# Make it a list
illegal_variable_names = illegal_variable_names_.replace(" ","").split(",")


def find_model_dat():
    """
    Find the file containing the definition of all the models in Xspec
    (model.dat) and return its path
    """

    # model.dat is in $HEADAS/../spectral

    headas_env = os.environ.get("HEADAS")

    assert headas_env is not None, ("You need to setup the HEADAS variable before importing this module."
                                    " See Heasoft documentation.")

    # Expand all variables and other things like ~
    headas_env = os.path.expandvars(os.path.expanduser(headas_env))

    # Lazy check that it exists

    assert os.path.exists(headas_env), "The HEADAS env. variable point to a non-existent directory: %s" % (headas_env)

    # Get one directory above HEADAS (i.e., $HEADAS/..)

    inferred_path = os.path.dirname(headas_env)

    # Now model.dat should be in $HEADAS/../spectral/manager

    final_path = os.path.join(inferred_path, 'spectral', 'manager', 'model.dat')

    # Check that model.dat exists

    assert os.path.exists(final_path), "Cannot find Xspec model definition file %s" % (final_path)

    return os.path.abspath(final_path)


def get_models(model_dat_path):
    """
    Parse the model.dat file from Xspec and returns a dictionary containing the definition of all the models

    :param model_dat_path: the path to the model.dat file
    :return: dictionary containing the definition of all XSpec models
    """

    with open(model_dat_path) as f:

        # model.dat is a text file, no size issues here (will fit in memory)

        model_dat = f.read()

    # Replace \r or \r\n with \n (the former two might be in the file if
    # it originates from Windows)
    if "\n" in model_dat:

        model_dat = model_dat.replace("\r", "")

    else:

        model_dat = model_dat.replace("\r", "\n")

    # Loop through the file and build the model definition
    # dictionary

    lines = model_dat.split("\n")

    model_definitions = collections.OrderedDict()

    for line in lines:

        match = re.match('''(.+(add|mul|con|acn).+)''', line)

        if match is not None:

            # This is a model definition

            tokens = line.split()

            if len(tokens) == 7:

                (model_name, n_parameters,
                 min_energy, max_energy,
                 library_function,
                 model_type,
                 flag) = line.split()

            else:

                (model_name, n_parameters,
                 min_energy, max_energy,
                 library_function,
                 model_type,
                 flag,
                 flag_2) = line.split()

            this_model = collections.OrderedDict()

            this_model['description'] = 'The %s model from XSpec (https://heasarc.gsfc.nasa.gov/xanadu/' \
                                                   'xspec/manual/XspecModels.html)' % model_name

            this_model['parameters'] = collections.OrderedDict()

            model_definitions[(model_name, library_function, model_type)] = this_model

        else:

            # This is a parameter definition

            if len(line.split()) == 0:
                # Empty line
                continue

            # Parameters are free by default, unless they have delta < 0
            # they are switch parameters or scale parameters

            free = True

            if line[0] == '$':

                # Probably a switch parameter

                free = False

                tokens = line.split()

                if len(tokens) == 2:

                    par_name = tokens[0][1:]

                    default_value = tokens[1]

                    par_unit = ""

                    hard_minimum, soft_minimum, soft_maximum, hard_maximum = (0, 0, 1, 1)

                elif len(tokens) == 3:

                    par_name = tokens[0][1:]

                    default_value = tokens[2]

                    par_unit = ""

                    hard_minimum, soft_minimum, soft_maximum, hard_maximum = (0, 0, 1, 1)

                else:

                    match = re.match('(\S+)\s+(\".+\"|[a-zA-Z]+)?(.+)*', line[1:])

                    par_name, par_unit, par_spec = match.groups()

                    tokens = par_spec.split()

                    if len(tokens) == 1:

                        default_value = tokens[0]
                        par_unit = ""
                        hard_minimum, soft_minimum, soft_maximum, hard_maximum = (0, 0, 1, 1)

                    else:

                        par_unit = ""

                        (default_value,
                         hard_minimum, soft_minimum,
                         soft_maximum, hard_maximum,
                         delta) = par_spec.split()


            else:

                # A normal parameter

                match = re.match('(\S+)\s+(\".+\"|\S+)(.+)', line)

                if match is None:

                    raise RuntimeError("Cannot parse parameter %s" % line)

                par_name, par_unit, par_spec = match.groups()

                if par_name[0] == '*':

                    # Scale parameter (always frozen)

                    par_name = par_name[1:]

                    free = False

                    tokens = par_spec.split()

                    if len(tokens) == 1:

                        default_value = tokens[0]
                        (hard_minimum, soft_minimum,
                         soft_maximum, hard_maximum,
                         delta) = (None, None, None, None, 0.1)

                    else:

                        (default_value,
                         hard_minimum, soft_minimum,
                         soft_maximum, hard_maximum,
                         delta) = par_spec.split()

                        delta = abs(float(delta))

                else:

                    (default_value,
                     hard_minimum, soft_minimum,
                     soft_maximum, hard_maximum,
                     delta) = par_spec.split()

                    delta = float(delta)

                    if delta <= 0:
                        free = False

                        delta = abs(delta)

            # Now fix the parameter name removing illegal characters
            # (for example 'log(A)' is not a legal name)

            par_name = re.sub('[\(,\)]', '_', par_name)
            par_name = re.sub('<', '_less_', par_name)
            par_name = re.sub('>', '_more_', par_name)
            par_name = re.sub('/', '_div_', par_name)
            par_name = re.sub('\-', '_minus_', par_name)
            par_name = re.sub('\+', '_plus_', par_name)
            par_name = re.sub('\.', '_dot_', par_name)
            par_name = re.sub('@', '_at_', par_name)

            # Parameter names must be lower case
            par_name = par_name.lower()

            # Some parameters are enclosed between ", like "z"
            par_name = par_name.replace('"', '')

            # "z" is a protected name in astromodels.
            if par_name == "z":

                par_name = "redshift"

            # Check that the parameter name is not an illegal Python name
            if par_name in illegal_variable_names:

                par_name = "xs_%s" % par_name

            # Sometimes the unit is " " which is not recognized by astropy
            if par_unit:

                par_unit = par_unit.replace("\"",'')

            # There are some errors in model.dat , like KeV instead of keV, and ergcm/s instead of erg cm /s
            # Let's correct them
            if par_unit == "KeV":

                par_unit = "keV"

            elif par_unit == "ergcm/s":

                par_unit = "erg cm / s"

            elif par_unit == "days":

                par_unit = "day"

            elif par_unit == "z" and par_name.lower() == "redshift":

                par_unit = ""

            # There are funny units in model.dat, like "Rs" (which means Schwarzschild radius) or other things
            # so let's try to convert the par_unit into an astropy.Unit instance. If that fails, use a unitless unit
            try:

                _ = u.Unit(par_unit)

            except ValueError:

                # Unit not recognized

                #warnings.warn("Unit %s is not recognized by astropy." % par_unit)

                par_unit = ''

            # Make sure that this is a valid python identifier
            # by matching it with the relative regexp
            if re.match('([a-zA-Z_][a-zA-Z0-9_]*)$', par_name) is None:

                raise ValueError("Illegal identifier name %s" % (par_name))

            this_model['parameters'][par_name] = {'initial value': float(default_value),
                                                  'desc': '(see https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/'
                                                          'XspecModels.html)',
                                                  'min': hard_minimum,
                                                  'max': hard_maximum,
                                                  'delta': float(delta),
                                                  'unit': par_unit,
                                                  'free': free}

    return model_definitions

class_definition_code = '''

class XS_$MODEL_NAME$(Function1D):

    """
$DOCSTRING$
    """

    __metaclass__ = FunctionMeta

    def _setup(self):

        # Link to the Xspec function
        self._model = _xspec.$XSPEC_FUNCTION$
        self._model_type = '$MODEL_TYPE$'

        if self._model_type == 'add':

            self._scale = 1e6

        else:

            # For multiplicative models, there is no differentiation to be made

            self._scale = 10

    def evaluate(self, x, $PARAMETERS_NAMES$):

        # Create a tuple of the current values of the parameters
        parameters_tuple = ($PARAMETERS_NAMES$,)

        # Finite difference differentiation of the Xspec
        # function for additive model. Indeed, xspec function return the integral
        # of the function on energy ranges, while we need the
        # differential flux. For multiplicative models instead Xspec returns just
        # the value for the multiplicative factor at the average between emin and
        # emax

        # Adapt the epsilon to the value to reduce the error

        epsilon = x / self._scale

        # In the normal case, when x is an array of more than one
        # element, the first call will succeed. If however there
        # is only one element in the array xspec will complain,
        # so handle that as a special case

        try:

            val = self._model(parameters_tuple, x - epsilon, x + epsilon)

        except TypeError:

            assert x.shape[0]==1, "This is a bug, xspec call failed and x is not only one element"
            
            val = self._model(parameters_tuple, ((x - epsilon)[0], (x + epsilon)[0]))[0]

        # val is now F(x-epsilon,x+epsilon) ~ f(x) * ( 2 * epsilon )

        if self._model_type == 'add':

            # In a additive model the function returns the integral over the bins

            return val / (2 * epsilon)

        else:

            # In a multiplicative model the function returns the average factor over
            # the bins

            return val

    def _set_units(self, x_unit, y_unit):

        # Parameters have units already
        pass

    def _integral(self, low_bounds, hi_bounds, $PARAMETERS_NAMES$):

        # Create a tuple of the current values of the parameters
        parameters_tuple = ($PARAMETERS_NAMES$,)

        return self._model(parameters_tuple, low_bounds, hi_bounds)

'''


def xspec_model_factory(model_name, xspec_function, model_type, definition):

    # If this is an additive model (model_type == 'add') we need to add
    # one more parameter (normalization)

    if model_type == 'add':
        
        definition['parameters']['norm'] = {'initial value': 1.0,
                                                  'desc': '(see https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/'
                                                          'XspecModels.html)',
                                                  'min': 0,
                                                  'max': None,
                                                  'delta': 0.1,
                                                  'unit': 'keV / (cm2 s)',
                                                  'free': True}

    assert model_type != 'con', "Convolution models are not yet supported"

    # Get a list of the parameter names
    parameters_names = ", ".join(definition['parameters'].keys())

    # Create the docstring
    docstring = my_yaml.dump(definition)

    # Create the class by substituting in the class_definition_code the
    # relevant things for this model

    code = class_definition_code.replace('$MODEL_NAME$', model_name)
    code = code.replace('$DOCSTRING$', docstring)
    code = code.replace('$PARAMETERS_NAMES$', parameters_names)
    code = code.replace('$XSPEC_FUNCTION$', xspec_function)
    code = code.replace('$MODEL_TYPE$', model_type)

    # Compile the code and create the class

    exec(code)

    # Return the class we just created

    class_name = 'XS_%s' % model_name

    return class_name, locals()[class_name]


def setup_xspec_models():

    classes = []

    if has_xspec:

        sys.stdout.write("Loading xspec models...")

        all_models = get_models(find_model_dat())

        for (model_name, xspec_function, model_type) in all_models:

            if model_type == 'con':

                # convolution models are not supported
                continue

            this_model = all_models[(model_name, xspec_function, model_type)]

            # When the class is created it is registered among the known functions in the function module
            # (it happens in the metaclass), so we don't need to do anything special here after the
            # class type is created

            this_class_name, this_class = xspec_model_factory(model_name, xspec_function, model_type, this_model)

            classes.append(this_class_name)



        sys.stdout.write("done\n")

    else:

        warnings.warn("XSpec is not available.", XSpecNotAvailable)

    return classes

# This will either work or issue a warning if XSpec is not available

new_functions = setup_xspec_models()

# Now import the new classes in the local namespace (if any)
# This is needed to make the classes pickeable

__all__ = []

for function_name in new_functions:

    __all__.append(function_name)

    locals()[function_name] = get_function_class(function_name)



__author__ = 'giacomov'

import collections

from astromodels.utils.io import display
from astromodels.dual_access_class import DualAccessClass


class DuplicatedNode(Exception):
    pass


def name_is_legal(name):
    """
    Check that the name is legal, i.e., it does not contain special characters nor spaces

    :param name: the name to check
    :return: True or False
    """

    # Define the set of valid characters
    valid_char_sequence = 'qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890-+_'

    valid = set(valid_char_sequence).issuperset(name)

    if valid:

        # Check that the name does not start with -
        if name[0] == '-':

            return False

    return valid


class Node(DualAccessClass):

    def __init__(self, name):

        self.__children = collections.OrderedDict()
        self.__parent = None

        assert name_is_legal(name),"Illegal characters in name %s. You can only use letters and numbers, " \
                                   "and + and - (but the name cannot start with -)" % name

        self._name = name

        super(Node, self).__init__('node', self.__children)

    @property
    def name(self):
        """
        Returns the name of the node

        :return: a string containing the name
        """
        return self._name

    @property
    def _children(self):
        return self.__children

    @property
    def path(self):

        return ".".join(self._get_path())

    def _reset_node(self):

        self.__children = collections.OrderedDict()
        self.__parent = None

    def _add_children(self, children):

        for child in children:

            self._add_child(child)

    def _add_child(self, new_child, name=None):

        new_child._set_parent(self)

        if name is None:

            name = new_child.name

        if name in self.__children:

            raise DuplicatedNode("You cannot use the same name (%s) for different nodes" % name)

        self.__children[name] = new_child

        # Add also an attribute with the name of the new child, to allow access with a syntax like
        # node.child

        self._add_attribute(name, new_child)

    def _get_child(self, child_name):

        return self.__children[child_name]

    def _remove_child(self, child_name):

        return self._del_attribute(child_name)

    def _set_parent(self, parent):

        self.__parent = parent

    def _get_parent(self):

        return self.__parent

    def _get_child_from_path(self, path):
        """
        Return a children below this level, starting from a path of the kind "this_level.something.something.name"

        :param path: the key
        :return: the child
        """

        keys = path.split(".")

        this_child = self

        for key in keys:

            try:

                this_child = this_child.get_child(key)

            except KeyError:

                raise KeyError("Child %s not found" % path)

        return this_child

    def _get_path(self):

        parent_names = []

        current_node = self

        while True:

            this_parent = current_node._get_parent()

            if this_parent is None:

                break

            else:

                parent_names.append(current_node.name)

                current_node = this_parent

        return parent_names[::-1]

    def to_dict(self, minimal=False):

        this_dict = collections.OrderedDict()

        for key, val in self.__children.iteritems():

            this_dict[key] = val.to_dict(minimal)

        return this_dict

    def _repr__base(self, rich_output):

        raise NotImplementedError("You should implement the __repr__base method for each class")

    def __repr__(self):
        """
        Textual representation for console

        :return: representation
        """

        return self._repr__base(rich_output=False)

    def _repr_html_(self):
        """
        HTML representation for the IPython notebook

        :return: HTML representation
        """

        return self._repr__base(rich_output=True)

    def display(self):
        """
        Display information about the point source.

        :return: (none)
        """

        # This will automatically choose the best representation among repr and repr_html

        display(self)

    def _find_instances(self, cls):
        """
        Find all the instances of cls below this node.

        :return: a dictionary of instances of cls
        """

        instances = collections.OrderedDict()

        for child_name, child in self._children.iteritems():

            if isinstance(child, cls):

                key_name = ".".join(child._get_path())

                instances[key_name] = child

                # Now check if the instance has children,
                # and if it does go deeper in the tree

                # NOTE: an empty dictionary evaluate as False

                if child._children:

                    instances.update(child._find_instances(cls))

            else:

                instances.update(child._find_instances(cls))

        return instances



__author__ = 'Giampaolo Rodola <g.rodola [AT] gmail [DOT] com>'

"""
Return disk usage statistics about the given path as a (total, used, free)
namedtuple.  Values are expressed in bytes.
"""
# Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
# License: MIT

import os
import collections

_ntuple_diskusage = collections.namedtuple('usage', 'total used free')

if hasattr(os, 'statvfs'):  # POSIX
    def disk_usage(path):
        st = os.statvfs(path)
        free = st.f_bavail * st.f_frsize
        total = st.f_blocks * st.f_frsize
        used = (st.f_blocks - st.f_bfree) * st.f_frsize
        return _ntuple_diskusage(total, used, free)

elif os.name == 'nt':  # Windows
    import ctypes
    import sys


    def disk_usage(path):
        _, total, free = ctypes.c_ulonglong(), ctypes.c_ulonglong(), ctypes.c_ulonglong()
        if sys.version_info >= (3,) or isinstance(path, unicode):
            fun = ctypes.windll.kernel32.GetDiskFreeSpaceExW
        else:
            fun = ctypes.windll.kernel32.GetDiskFreeSpaceExA
        ret = fun(path, ctypes.byref(_), ctypes.byref(total), ctypes.byref(free))
        if ret == 0:
            raise ctypes.WinError()
        used = total.value - free.value
        return _ntuple_diskusage(total.value, used, free.value)
else:
    raise NotImplementedError("platform not supported")

disk_usage.__doc__ = __doc__



from astromodels.parameter import Parameter
from astromodels.my_yaml import my_yaml
from astromodels.utils.pretty_list import dict_to_list
from astromodels.tree import Node
from astromodels.utils.table import dict_to_table
from astromodels.units import get_units
import astropy.units as u
import functools


import collections
import warnings
import sys
import os
import copy
import uuid
import re
import ast
from yaml.reader import ReaderError
import numpy as np
import scipy.integrate
import inspect

__author__ = 'giacomov'


try:

    from IPython.display import display, HTML

except:

    has_ipython = False

else:

    has_ipython = True


class WarningNoTests(ImportWarning):
    pass


class FunctionDefinitionError(Exception):
    pass


class DesignViolation(Exception):
    pass


class ModelAssertionViolation(Exception):
    pass


class WrongDimensionality(Exception):
    pass


class TestSpecificationError(Exception):
    pass


class TestFailed(Exception):
    pass


class DocstringIsNotRaw(ValueError):
    pass


class UnknownFunction(ValueError):
    pass


class UnknownParameter(ValueError):
        pass


# Value to indicate that no latex formula has been given
NO_LATEX_FORMULA = '(no latex formula available)'

# Codes to indicate to Composite Function the operation between two functions
_operations = {'+': np.add,
               '-': np.subtract,
               '*': np.multiply,
               '/': np.divide,
               '**': np.power,
               'abs': np.abs,
               'of': 'compose'}


# This dictionary will contain the known function by name, so that the model_parser can instance
# them by looking into this dictionary. It will be filled by the FunctionMeta meta-class.

_known_functions = {}


def memoize(method):
    """
    A decorator for the 2d functions which memoize the results of 1 call (useful when the minimizer is taking partial
    derivatives and calls the 2d function several times with the same arguments)

    :param method: method to be memoized
    :return: the decorated method
    """

    # Need more testing for this
    # TODO: test memoization

    # return method

    cache = method.cache = collections.OrderedDict()

    @functools.wraps(method)
    def memoizer(instance, x, *args, **kwargs):

        # Create a tuple because a tuple is hashable

        unique_id = tuple(x.value for x in instance.parameters.values()) + (x.shape[0],x.min(),x.max())

        # Create a unique identifier for this combination of inputs

        key = hash(unique_id)

        # Let's do it this way so we only look into the dictionary once

        result = cache.get(key)

        if result is not None:

            return result

        else:

            result = method(instance, x, *args, **kwargs)

            cache[key] = result

            if len(cache) > 10:

                # Remove the 10 elements that were put in first

                [cache.popitem(False) for i in range(5)]

            return result

    # Add the function as a "attribute" so we can access it
    memoizer.input_object = method

    return memoizer


# The following is a metaclass for all the functions
class FunctionMeta(type):
    """
    A metaclass for the models, which takes care of setting up the parameters and the other attributes
    according to the definition given in the documentation of the function class.
    """

    def __init__(cls, name, bases, dct):

        # This method is called at the very beginning during the import of the module,
        # after the class cls has been built.
        # We use it to perform simple checks on the function class to verify that it is
        # implemented well, and add properties and methods according to the definition
        # given in the docstring of cls

        # Enforce the presence of the evaluate method

        if 'evaluate' not in dct:

            raise AttributeError("You have to implement the 'evaluate' method in %s" % name)

        # Minimal check of the 'evaluate' function

        variables, parameters_in_calling_sequence = FunctionMeta.check_calling_sequence(name, 'evaluate',
                                                                                        cls.evaluate, ['x', 'y', 'z'])

        # We also need the method _set_units
        if '_set_units' not in dct:

            raise AttributeError("You have to implement the '_set_units' method in %s" % name)

        #print("Function %s: \n variables: %s\n parameters: %s\n" % (cls.__name__, ",".join(variables),
        #                                                            ",".join(parameters_in_calling_sequence)))

        # Figure out the dimensionality of this function

        n_dim = len(variables)

        # Store the dimensionality in the *type*

        cls._n_dim = n_dim

        # Now parse the documentation of the function which contains the parameter specification

        # The doc is a YAML document containing among other things the definition of the parameters

        # Parse it

        try:

            function_definition = my_yaml.load(cls.__doc__)

        except ReaderError:

            raise DocstringIsNotRaw("Docstring parsing has failed. "
                                    "Did you remember to specify the docstring of %s as raw? "
                                    "To do that, you have to put a r before the docstring, "
                                    '''like in \n\nr"""\n(docstring)\n"""\n\ninstead of just\n\n'''
                                    '''"""\ndocstring\n"""''' % name)

        else:

            # Store the function definition in the type

            cls._function_definition = function_definition

        # Enforce the presence of a description and of a parameters dictionary

        assert "description" in function_definition.keys(), "You have to provide a 'description' token in the " \
                                                            "documentation of class %s" % name

        assert "parameters" in function_definition.keys(), "You have to provide a 'parameters' token in the " \
                                                           "documentation of class %s" % name

        # If there is a latex formula, store it in the type

        if 'latex' in function_definition:

            # First remove the escaping we did to overcome the limitation of the YAML parser

            latex_formula = function_definition['latex'].replace(r"\\", chr(92))

        else:

            latex_formula = NO_LATEX_FORMULA

        # Store latex formula in the type
        cls._latex = latex_formula

        # Store the name in the type
        cls._name = cls.__name__

        # Parse the parameters' dictionary
        assert isinstance(function_definition['parameters'], dict), "Wrong syntax in 'parameters' token. It must be " \
                                                                    "a dictionary. Refers to the documentation."

        # Add the parameters as attribute of the *type*. During the __call__ method below this dictionary will be used
        # to create a copy of each parameter which will be made available as attribute of the instance.

        cls.__parameters = collections.OrderedDict()

        for parameter_name, parameter_definition in function_definition['parameters'].iteritems():
            this_parameter = FunctionMeta.parse_parameter_definition(name, parameter_name, parameter_definition)

            cls.__parameters[this_parameter.name] = this_parameter

        # Now check that all the parameters used in 'evaluate' are part of the documentation,
        # and that there are no unused parameters

        set1 = set(cls.__parameters.keys())
        set2 = set(parameters_in_calling_sequence)

        if set1 != set2:

            # The parameters are different. Figure out who is missing and raise an exception accordingly

            if set1 > set2:

                missing = set1 - set2

                msg = "Parameters %s have init values but are not used in 'evaluate' in %s" % (",".join(missing), name)

            else:

                missing = set2 - set1

                msg = "Parameters %s are used in 'evaluate' but do not have init values in %s" % \
                      (",".join(missing), name)

            raise FunctionDefinitionError(msg)

        # Now add the constructor to the class
        cls.__init__ = FunctionMeta.class_init

        # Finally, add the info() method to the type so that it can be called even without instancing the class

        def info():

            repr_dict = collections.OrderedDict()

            repr_dict['description'] = function_definition['description']

            if 'latex' in function_definition:
                repr_dict['formula'] = function_definition['latex']

            # Add the description of each parameter and their current value
            repr_dict['default parameters'] = collections.OrderedDict()

            for parameter_name in cls.__parameters.keys():

                repr_dict['default parameters'][parameter_name] = cls.__parameters[parameter_name].to_dict()

            if has_ipython:

                display(HTML(dict_to_list(repr_dict, html=True)))

            else:

                print(dict_to_list(repr_dict, html=False))

        cls.info = staticmethod(info)

        # All went well, add this as a known function

        _known_functions[name] = cls

        # Finally call the type init

        super(FunctionMeta, cls).__init__(name, bases, dct)

    @staticmethod
    def check_calling_sequence(name, function_name, function, possible_variables):
        """
        Check the calling sequence for the function looking for the variables specified.
        One or more of the variables can be in the calling sequence. Note that the
        order of the variables will be enforced.
        It will also enforce that the first parameter in the calling sequence is called 'self'.

        :param function: the function to check
        :param possible_variables: a list of variables to check, The order is important, and will be enforced
        :return: a tuple containing the list of found variables, and the name of the other parameters in the calling
        sequence
        """

        # Get calling sequence

        # If the function has been memoized, it will have a "input_object" member

        try:

            calling_sequence = inspect.getargspec(function.input_object).args

        except AttributeError:

            # This might happen if the function is with memoization

            calling_sequence = inspect.getargspec(function).args

        assert calling_sequence[0] == 'self', "Wrong syntax for 'evaluate' in %s. The first argument " \
                                              "should be called 'self'." % name

        # Figure out how many variables are used

        variables = filter(lambda var: var in possible_variables, calling_sequence)

        # Check that they actually make sense. They must be used in the same order
        # as specified in possible_variables

        assert len(variables) > 0, "The name of the variables for 'evaluate' in %s must be one or more " \
                                   "among %s, instead of %s" % (name, ','.join(possible_variables), ",".join(variables))

        if variables != possible_variables[:len(variables)]:
            raise AssertionError("The variables %s are out of order in '%s' of %s. Should be %s."
                                 % (",".join(variables), function_name, name, possible_variables[:len(variables)]))

        other_parameters = filter(lambda var: var not in variables and var != 'self', calling_sequence)

        return variables, other_parameters

    @staticmethod
    def parse_parameter_definition(func_name, par_name, definition):

        # Parse definition of parameter

        # Enforce the presence of attributes 'value' and 'desc'

        if 'initial value' not in definition:
            raise FunctionDefinitionError("Error for parameter %s of function %s: value for parameter must be"
                                          " specified" % (par_name, func_name))

        if 'desc' not in definition:
            raise FunctionDefinitionError("Error for parameter %s of function %s: desc for parameter must be"
                                          " specified" % (par_name, func_name))

        # Fetch attributes

        # Use unitless parameters when building the function, if no unit is specified, otherwise
        # use that unit
        if 'unit' not in definition or definition['unit'] is None or definition['unit'] == '':

            du = u.dimensionless_unscaled

        else:

            du = u.Unit(definition['unit'])

        def _parse_value(val):

            if val is None:

                return None

            if isinstance(val, str):

                return eval(val)

            else:

                return float(val)

        value = _parse_value(definition['initial value'])
        desc = definition['desc']

        # Optional attributes are either None if not specified, or the value specified

        min_value = (None if 'min' not in definition else _parse_value(definition['min']))
        max_value = (None if 'max' not in definition else _parse_value(definition['max']))
        delta = (None if 'delta' not in definition else _parse_value(definition['delta']))
        unit = du

        # A parameter can be fixed by using fix=yes, otherwise it is free by default

        free = (True if 'fix' not in definition else not bool(definition['fix']))

        new_parameter = Parameter(par_name, value, min_value=min_value, max_value=max_value,
                                  delta=delta, desc=desc, free=free, unit=unit)

        return new_parameter

    @staticmethod
    def class_init(instance, **kwargs):

        # Create a copy of the parameters dictionary which is in the type,
        # otherwise every instance would share the same dictionary

        copy_of_parameters = collections.OrderedDict()

        # Fill it by duplicating the parameters contained in the dictionary in the type

        for key, value in type(instance).__parameters.iteritems():

            copy_of_parameters[key] = value.duplicate()

            # If the user has specified a value in the constructor, update the
            # corresponding parameters value. This allow to use a constructor as:
            # my_powerlaw = powerlaw(logK=1.0, index=-2)

            if key in kwargs:

                copy_of_parameters[key].value = kwargs[key]

        # Now check that all the parameters specified in the kwargs are actually parameters of this function
        for key in kwargs.keys():

            try:

                copy_of_parameters[key]

            except KeyError:

                raise UnknownParameter("You specified an init value for %s, which is not a "
                                       "parameter of function %s" % (key, type(instance)._name))

        # Now call the parent class
        n_dim = type(instance)._n_dim

        if n_dim == 1:

            Function1D.__init__(instance,
                              type(instance)._name,
                              type(instance)._function_definition,
                              copy_of_parameters)

        elif n_dim == 2:

            Function2D.__init__(instance,
                                type(instance)._name,
                                type(instance)._function_definition,
                                copy_of_parameters)

        elif n_dim == 3:

            Function3D.__init__(instance,
                                type(instance)._name,
                                type(instance)._function_definition,
                                copy_of_parameters)

        # Last, if the class provides a setup method, call it
        if hasattr(instance, "_setup"):

            instance._setup()


class Function(Node):

    def __init__(self, name, function_definition, parameters):

        # (this is called by the constructor defined in the metaclass)

        # Store name, number of dimensions and the latex formula

        # Store also the function definition

        assert 'description' in function_definition,"Function definition must contain a description"

        if 'latex' not in function_definition:

            function_definition['latex'] = '$n.a.$'

        self._function_definition = function_definition

        # Set up the node

        Node.__init__(self, name)

        # Add the parameters as children. Since the name of the key in the dictionary might
        # be different than the actual name of the parameter, use the .add_child method instead
        # of the add_children method

        for child_name, child in parameters.iteritems():

            self._add_child(child, child_name)

        # Now generate a unique identifier (UUID) in a thread safe, multi-processing safe
        # way. This is used for example in the CompositeFunction class to keep track of the different
        # instances of the same function
        self._uuid = "{" + str(self._generate_uuid()) + "}"

    @property
    def n_dim(self):

        return type(self)._n_dim

    @property
    def free_parameters(self):
        """
        Returns a dictionary of free parameters for this function

        :return: dictionary of free parameters
        """

        free_parameters = collections.OrderedDict([(k,v) for k, v in self.parameters.iteritems() if v.free])

        return free_parameters

    @staticmethod
    def _generate_uuid():
        """
        Generate a unique identifier for this function.

        :return: the UUID
        """
        return uuid.UUID(bytes=os.urandom(16), version=4)

    @property
    def description(self):
        """
        Returns a description for this function
        """

        return self._function_definition['description']

    # Add a property returning the parameters dictionary
    @property
    def parameters(self):
        """
        Returns a dictionary of parameters
        """
        return self._children

    @property
    def latex(self):
        """
        Returns the LaTEX formula for this function
        """
        return self._function_definition['latex']

    # Define now all the operators which allow to combine functions. Each operator will return a new
    # instance of a CompositeFunction, which can then be used as a function on its own

    @staticmethod
    def __get_second_uuid(other_instance):
        """
        Return a name for the object. If the object is a function instance, return its name. Otherwise, return
        the object itself.

        :param other_instance:
        :return:
        """
        if hasattr(other_instance,'uuid'):

            # Another function

            second_uuid = other_instance.uuid

        else:

            # A number

            second_uuid = '%s' % other_instance

        return second_uuid

    def of(self, another_function):
        """
        Compose this function with another as in this_function(another_function(x))

        :param another_function: another function to compose with the current one
        :return: a composite function instance
        """
        return CompositeFunction('of', self, another_function)

    def __neg__(self):

        return CompositeFunction('-', self)

    def __abs__(self):

        return CompositeFunction('abs', self)

    def __pow__(self, other_instance):

        second_uuid = self.__get_second_uuid(other_instance)

        return CompositeFunction('**', self, other_instance)

    def __rpow__(self, other_instance):

        return CompositeFunction('**', other_instance, self)

    def __add__(self, other_instance):
        """
        Return a composite function where the current instance is summed with the given instance

        :param other_instance: the other instance. This can be either a number, or a Function instance.
        """

        return CompositeFunction('+', self, other_instance)

    __radd__ = __add__

    def __sub__(self, other_instance):

        return CompositeFunction('-', self, other_instance)

    __rsub__ = __sub__

    def __mul__(self, other_instance):

        return CompositeFunction('*', self, other_instance)

    __rmul__ = __mul__

    def __div__(self, other_instance):

        return CompositeFunction('/', self, other_instance)

    def __rdiv__(self, other_instance):

        return CompositeFunction('/', other_instance, self)

    __truediv__ = __div__
    __rtruediv__ = __rdiv__

    def _repr__base(self, rich_output):

        repr_dict = collections.OrderedDict()

        repr_dict['description'] = self._function_definition['description']

        if 'latex' in self._function_definition:

            repr_dict['formula'] = self._function_definition['latex']

        # Add the description of each parameter and their current value
        repr_dict['parameters'] = collections.OrderedDict()

        for parameter_name in self._children.keys():

            repr_dict['parameters'][parameter_name] = self._children[parameter_name].to_dict()

        return dict_to_list(repr_dict, rich_output)

    @property
    def uuid(self):
        """
        Returns the ID of the current function. The ID is used by the CompositeFunction class to keep track of the
        unique instances of each function. It should not be used by the user for any specific purpose.

        :return: (none)
        """
        return self._uuid

    def duplicate(self):
        """
        Create a copy of the current function with all the parameters equal to the current value

        :return: a new copy of the function
        """

        # Create a copy

        function_copy = copy.deepcopy(self)

        return function_copy

    def get_boundaries(self):
        """
        Returns the boundaries of this function. By default there is no boundary, but subclasses can
        override this.

        :return: a tuple of tuples containing the boundaries for each coordinate, or None if there are no boundaries
        """

        return None


class Function1D(Function):

    def __init__(self, name, function_definition, parameters):

        Function.__init__(self, name, function_definition, parameters)

        self._x_unit = None
        self._y_unit = None

    def get_wrapper(self):
        """
        Returns a python function which can be used to call this function with the parameters in the calling sequence.
        In other words, if you can call this function with f(x), with the returned wrapper you can call it with
        f(x, parameter1, parameter2...), where parameter1, parameter2... are the *free* parameters.

        :return: a python function
        """

        # Build a list of free parameters
        free_parameters = [k for k, v in self.parameters.iteritems() if v.free]

        # Prepare the method to set the parameters to their current value

        def set_parameters(*args):

            for i in range(len(args)):
                self._get_child(free_parameters[i]).value = args[i]

        # Prepare the variable description

        variables = None

        if self.n_dim == 1:

            variables = 'x'

        elif self.n_dim == 2:

            variables = 'x,y'

        elif self.n_dim == 3:

            variables = 'x,y,z'

        # Prepare the free parameters string

        free_parameters_string = ",".join(free_parameters)

        # Build some code to generate a wrapper which takes care of updating the value of the parameters
        # and return the value of the function

        wrapper_code = '''

        def wrapper(%s, %s):

            set_parameters(%s)

            return self(%s)

        ''' % (variables, free_parameters_string, free_parameters_string, variables)

        # print(wrapper_code)

        exec (wrapper_code.replace("        ", "")) in locals()

        return wrapper

    def evaluate(self, x, *args, **kwargs):

        raise NotImplementedError("You have to re-implement this")

    def set_units(self, in_x_unit, in_y_unit):

        try:

            in_x_unit = u.Unit(in_x_unit)
            in_y_unit = u.Unit(in_y_unit)

        except:

            raise TypeError("Could not get a Unit instance from provided units when setting units "
                            "for function %s" % self.name)

        self._x_unit = in_x_unit
        self._y_unit = in_y_unit

        # Now call the underlying method to set units, which is defined by each function
        self._set_units(self._x_unit, self._y_unit)

    def _set_units(self, x_unit, y_unit):

        # This will be overridden by derived classes

        raise NotImplementedError("You have to implement the method _set_units for function %s" % self.name)

    @property
    def x_unit(self):
        return self._x_unit

    @property
    def y_unit(self):
        return self._y_unit

    def __call__(self, x, *args, **kwargs):

        # This method's code violates explicitly duck typing. The reason is that astropy.units introduce a very
        # significant overload on any computation. For this reason we treat differently the case with units from
        # the case without units, so that the latter case remains fast. Also, transforming an input
        # which is not an array into an array introduce a significant overload (10 microseconds or so), so we perform
        # this transformation only when strictly required

        if isinstance(x, np.ndarray):

            # We have an array as input

            if not isinstance(x, u.Quantity):

                # This is a normal array, let's use the fast call (without units)

                return self._call_without_units(x, *args, **kwargs)

            else:

                # This is an array with units, let's use the slow call which preserves units

                assert self.y_unit is not None, "In order to use units you need to use the function as a spectrum or " \
                                                "as something else," \
                                                "or you need to explicitly set the units."

                results = self._call_with_units(x, *args, **kwargs)

                # Now convert to the expected y unit
                return results.to(self.y_unit)

        else:

            # This is either a single number or a list
            if not isinstance(x, u.Quantity):

                # Transform the input to an array of floats

                new_input = np.array(x, dtype=float, ndmin=1, copy=False)

                # Compute the function

                result = self._call_without_units(new_input, *args, **kwargs)

                # Now remove all dimensions of size 1. For example, an array of shape (1,) will become a single number.

                return np.squeeze(result)

            else:

                # This is a single number with units, let's transform it to an array with units

                new_input = np.array(x, dtype=float, ndmin=1, copy=False) * x.unit

                # Compute the function with units

                result = self._call_with_units(new_input, *args, **kwargs)

                # Now remove all dimensions of size 1. For example, an array of shape (1,) will become a single number.
                # Let's also convert the result to the expected units

                return np.squeeze(result).to(self.y_unit)

    def _call_with_units(self, x, *args, **kwargs):

        # Gather the current parameters' values with units

        for parameter_name, parameter in self._children.iteritems():

            kwargs[parameter_name] = parameter.value * parameter.unit

        try:

            results = self.evaluate(x, *args, **kwargs)

        except u.UnitsError:

            raise

            raise u.UnitsError("Looks like you didn't provide all the units, or you provided the wrong ones, when "
                               "calling function %s" % self.name)

        else:

            return results

    def _call_without_units(self, x, *args, **kwargs):

        # Gather the current parameters' values without units, which means that the whole computation
        # will be without units, with a big speed gain (~10x)

        for parameter_name, parameter in self._children.iteritems():
            kwargs[parameter_name] = parameter.value

        return self.evaluate(x, *args, **kwargs)


class Function2D(Function):

    def __init__(self, name, function_definition, parameters):

        Function.__init__(self, name, function_definition, parameters)

        self._x_unit = None
        self._y_unit = None
        self._z_unit = None

    def evaluate(self, x, y, *args, **kwargs):

        raise NotImplementedError("You have to re-implement this")

    def set_units(self, in_x_unit, in_y_unit, in_z_unit):

        try:

            in_x_unit = u.Unit(in_x_unit)
            in_y_unit = u.Unit(in_y_unit)
            in_z_unit = u.Unit(in_z_unit)

        except:

            raise TypeError("Could not get a Unit instance from provided units when setting units "
                            "for function %s" % self.name)

        self._x_unit = in_x_unit
        self._y_unit = in_y_unit
        self._z_unit = in_z_unit

        # Now call the underlying method to set units, which is defined by each function
        self._set_units(self._x_unit, self._y_unit, self._z_unit)

    def _set_units(self, x_unit, y_unit, z_unit):

        # This will be overridden by derived classes

        raise NotImplementedError("You have to implement the method _set_units for function %s" % self.name)

    @property
    def x_unit(self):
        return self._x_unit

    @property
    def y_unit(self):
        return self._y_unit

    @property
    def z_unit(self):
        return self._z_unit

    def __call__(self, x, y, *args, **kwargs):

        # This method's code violates explicitly duck typing. The reason is that astropy.units introduce a very
        # significant overload on any computation. For this reason we treat differently the case with units from
        # the case without units, so that the latter case remains fast. Also, transforming an input
        # which is not an array into an array introduce a significant overload (10 microseconds or so), so we perform
        # this transformation only when strictly required

        assert type(x) == type(y), "You have to use the same type for x and y"

        if isinstance(x, np.ndarray):

            # We have an array as input

            if not isinstance(x, u.Quantity):

                # This is a normal array, let's use the fast call (without units)

                return self._call_without_units(x, y, *args, **kwargs)

            else:

                # This is an array with units, let's use the slow call which preserves units

                results = self._call_with_units(x, y, *args, **kwargs)

                # Now convert to the expected y unit
                return results.to(self.z_unit)

        else:

            # This is either a single number or a list
            if not isinstance(x, u.Quantity):

                # Transform the input to an array of floats

                new_x = np.array(x, dtype=float, ndmin=1, copy=False)
                new_y = np.array(y, dtype=float, ndmin=1, copy=False)

                # Compute the function

                result = self._call_without_units(new_x, new_y, *args, **kwargs)

                # Now remove all dimensions of size 1. For example, an array of shape (1,) will become a single number.

                return np.squeeze(result)

            else:

                # This is a single number with units, let's transform it to an array with units

                new_x = np.array(x, dtype=float, ndmin=1, copy=False) * x.unit
                new_y = np.array(y, dtype=float, ndmin=1, copy=False) * y.unit

                # Compute the function with units

                result = self._call_with_units(new_x, new_y, *args, **kwargs)

                # Now remove all dimensions of size 1. For example, an array of shape (1,) will become a single number.
                # Let's also convert the result to the expected units

                return np.squeeze(result).to(self.z_unit)

    def _call_with_units(self, x, y, *args, **kwargs):

        # Gather the current parameters' values with units

        for parameter_name, parameter in self._children.iteritems():
            kwargs[parameter_name] = parameter.value * parameter.unit

        try:

            results = self.evaluate(x, y, *args, **kwargs)

        except u.UnitsError:

            raise u.UnitsError("Looks like you didn't provide all the units, or you provided the wrong ones, when "
                               "calling function %s" % self.name)

        else:

            return results

    @memoize
    def _call_without_units(self, x, y, *args, **kwargs):

        # Gather the current parameters' values without units, which means that the whole computation
        # will be without units, with a big speed gain (~10x)

        for parameter_name, parameter in self._children.iteritems():
            kwargs[parameter_name] = parameter.value

        return self.evaluate(x, y, *args, **kwargs)


class Function3D(Function):

    def __init__(self, name, function_definition, parameters):

        Function.__init__(self, name, function_definition, parameters)

        self._x_unit = None
        self._y_unit = None
        self._z_unit = None
        self._w_unit = None

    def evaluate(self, x, y, z, *args, **kwargs):

        raise NotImplementedError("You have to re-implement this")

    def set_units(self, in_x_unit, in_y_unit, in_z_unit, in_w_unit):

        try:

            in_x_unit = u.Unit(in_x_unit)
            in_y_unit = u.Unit(in_y_unit)
            in_z_unit = u.Unit(in_z_unit)
            in_w_unit = u.Unit(in_w_unit)

        except:

            raise TypeError("Could not get a Unit instance from provided units when setting units "
                            "for function %s" % self.name)

        self._x_unit = in_x_unit
        self._y_unit = in_y_unit
        self._z_unit = in_z_unit
        self._w_unit = in_w_unit

        # Now call the underlying method to set units, which is defined by each function
        self._set_units(self._x_unit, self._y_unit, self._z_unit, self._w_unit)

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # This will be overridden by derived classes

        raise NotImplementedError("You have to implement the method _set_units for function %s" % self.name)

    @property
    def x_unit(self):
        return self._x_unit

    @property
    def y_unit(self):
        return self._y_unit

    @property
    def z_unit(self):
        return self._z_unit

    @property
    def w_unit(self):
        return self._w_unit

    def __call__(self, x, y, z, *args, **kwargs):

        # This method's code violates explicitly duck typing. The reason is that astropy.units introduce a very
        # significant overload on any computation. For this reason we treat differently the case with units from
        # the case without units, so that the latter case remains fast. Also, transforming an input
        # which is not an array into an array introduce a significant overload (10 microseconds or so), so we perform
        # this transformation only when strictly required

        assert type(x) == type(y) and type(y) == type(z), "You have to use the same type for x, y and z"

        if isinstance(x, np.ndarray) and x.shape != ():

            # We have an array as input

            if not isinstance(x, u.Quantity):

                # This is a normal array, let's use the fast call (without units)

                return self._call_without_units(x, y, z, *args, **kwargs)

            else:

                # This is an array with units, let's use the slow call which preserves units

                results = self._call_with_units(x, y, z, *args, **kwargs)

                # Now convert to the expected y unit
                return results.to(self.w_unit)

        else:

            # This is either a single number or a list
            if not isinstance(x, u.Quantity):

                # Transform the input to an array of floats

                new_x = np.array(x, dtype=float, ndmin=1, copy=False)
                new_y = np.array(y, dtype=float, ndmin=1, copy=False)
                new_z = np.array(z, dtype=float, ndmin=1, copy=False)

                # Compute the function

                result = self._call_without_units(new_x, new_y, new_z, *args, **kwargs)

                # Now remove all dimensions of size 1. For example, an array of shape (1,) will become a single number.

                return np.squeeze(result)

            else:

                # This is a single number with units, let's transform it to an array with units

                new_x = np.array(x, dtype=float, ndmin=1, copy=False) * x.unit
                new_y = np.array(y, dtype=float, ndmin=1, copy=False) * y.unit
                new_z = np.array(z, dtype=float, ndmin=1, copy=False) * z.unit

                # Compute the function with units

                result = self._call_with_units(new_x, new_y, new_z, *args, **kwargs)

                # Now remove all dimensions of size 1. For example, an array of shape (1,) will become a single number.
                # Let's also convert the result to the expected units

                return np.squeeze(result).to(self.w_unit)

    def _call_with_units(self, x, y, z, *args, **kwargs):

        # Gather the current parameters' values with units

        for parameter_name, parameter in self._children.iteritems():

            kwargs[parameter_name] = parameter.value * parameter.unit

        try:

            results = self.evaluate(x, y, z, *args, **kwargs)

        except u.UnitsError:

            raise u.UnitsError("Looks like you didn't provide all the units, or you provided the wrong ones, when "
                               "calling function %s" % self.name)

        else:

            return results

    def _call_without_units(self, x, y, z, *args, **kwargs):

        # Gather the current parameters' values without units, which means that the whole computation
        # will be without units, with a big speed gain (~10x)

        for parameter_name, parameter in self._children.iteritems():

            kwargs[parameter_name] = parameter.value

        return self.evaluate(x, y, z, *args, **kwargs)


class CompositeFunction(Function):

    def __init__(self, operation, function_or_scalar_1, function_or_scalar_2=None):

        assert operation in _operations,"Do not know operation %s" % operation

        # Set the new evaluate

        if function_or_scalar_2 is None:

            # Unary operation

            self.set_evaluate(self._composite_function_factory_unary(function_or_scalar_1, _operations[operation]))

        else:

            # Binary operation

            # Check if this is a function composition with the .of method
            if _operations[operation] == 'compose':

                # Can only compose functions of 1 variable

                assert hasattr(function_or_scalar_2, 'evaluate'), "Second member of .of cannot be a scalar"

                assert function_or_scalar_1.n_dim == 1 and function_or_scalar_2.n_dim == 1, "Can only compose " \
                                                                                            "with .of functions of " \
                                                                                            "1 variable"

                def new_evaluate(*args):

                    value = function_or_scalar_2(*args)

                    return function_or_scalar_1(value, *(args[1:]))

                self.set_evaluate(new_evaluate)

            else:

                self.set_evaluate(self._composite_function_factory_binary(function_or_scalar_1, _operations[operation],
                                                                          function_or_scalar_2))

        # Save a description, but using the unique IDs of the functions involved, to keep track
        # of where they appear in the expression

        self._uuid_expression = self._get_uuid_expression(operation, function_or_scalar_1, function_or_scalar_2)

        # Makes the list of unique functions which compose this composite function.

        self._functions = []

        for function in [function_or_scalar_1, function_or_scalar_2]:

            # Check whether this is already a composite function. If it is, add the functions contained
            # in it

            if isinstance(function, CompositeFunction):

                for sub_function in function.functions:

                    if sub_function not in self._functions:

                        self._functions.append(sub_function)

            elif isinstance(function, Function):

                # This is a simple function. Add it only if it is not there already (avoid duplicate)

                if function not in self._functions:

                    self._functions.append(function)

            else:

                # This is a scalar, no need to add it among the functions

                pass

        # Now assign a unique name to all the functions, to make clear which is which in the definition
        # and give an easy way for the user to understand which parameter belongs to which function

        self._id_to_uid = {}

        expression = self._uuid_expression

        for i,function in enumerate(self._functions):

            self._id_to_uid[i+1] = function.uuid

            expression = expression.replace(function.uuid, "%s{%s}" % (function.name, i+1))

        # Save the expression
        self._expression = expression

        # Build the parameters dictionary assigning a new name to each parameter to account for possible
        # duplicates.

        parameters = collections.OrderedDict()

        for i, function in enumerate(self._functions):

            for parameter_name, parameter in function.parameters.iteritems():

                # New name to avoid possible duplicates

                new_name = "%s_%i" % (parameter.name, i+1)

                # Store the parameter under the new name (obviously this is a reference to the
                # parameter, not a copy, as always in python)

                parameters[new_name] = parameter
                parameter._change_name(new_name)

        # Now build a meaningful description

        _function_definition = {'description': self.expression, 'latex': NO_LATEX_FORMULA}

        Function.__init__(self, 'composite', _function_definition, parameters)

        self._uuid = self._uuid_expression

    def set_units(self, x_unit, y_unit):

        # Just rely on the single functions to adjust themselves.

        for function in self.functions:

            function.set_units(x_unit, y_unit)

    @property
    def expression(self):
        return self._expression

    @staticmethod
    def _get_uuid_expression(operation, name_1, name_2=None):

        if name_2 is None:

            return '(%s %s)' % (operation, name_1.uuid)

        if hasattr(name_1, 'uuid'):

            name_1_uuid = name_1.uuid

        else:

            name_1_uuid = '%s' % name_1

        if hasattr(name_2,'uuid'):

            name_2_uuid = name_2.uuid

        else:

            name_2_uuid = '%s' % name_2

        return '(%s %s %s)' % (name_1_uuid, operation, name_2_uuid)

    @staticmethod
    def _composite_function_factory_unary(instance, numpy_operator):

        def new_evaluate(*args, **kwargs):

            return numpy_operator(instance.__call__(*args, **kwargs))

        return new_evaluate

    @staticmethod
    def _composite_function_factory_binary(first_instance, numpy_operator, second_instance):

        # Check whether the second member is a function, or a number

        if hasattr(first_instance, 'evaluate'):

            if hasattr(second_instance, 'evaluate'):

                # Get number of arguments for first function


                def new_evaluate(*args):

                    return numpy_operator(first_instance.__call__(*args),
                                          second_instance.__call__(*args))

                return new_evaluate

            else:

                def new_evaluate(*args):

                    return numpy_operator(first_instance.__call__(*args),
                                          second_instance)

                return new_evaluate

        else:

            if hasattr(second_instance, 'evaluate'):

                def new_evaluate(*args):

                    return numpy_operator(first_instance,
                                          second_instance.__call__(*args))

                return new_evaluate

            else:

                # Should never get here!

                raise RuntimeError("Should never get here")

    @property
    def functions(self):
        "A list containing the function used to build this composite function"
        return self._functions

    def evaluate(self):

        raise NotImplementedError("You cannot instance and use a composite function by itself. Use the factories.")

    def set_evaluate(self, new_evaluate_method):
        """
        This is called by the factory which create the composite function, and set the evaluate method
        to the new method which will collect the results from all the functions
        """

        self.evaluate = new_evaluate_method

    # Override the __call__ method of the Function class because the single functions in _functions
    # will handle their own collection of parameters

    def __call__(self, *args, **kwargs):

        return self.evaluate(*args, **kwargs)

    # Override the to_dict method of the Node class to add the expression to re-build this
    # composite function
    def to_dict(self, minimal=False):

        data = super(CompositeFunction, self).to_dict(minimal)

        if not minimal:

            data['expression'] = self._expression

        return data


def get_function(function_name, composite_function_expression=None):
    """
    Returns the function "name", which must be among the known functions or a composite function.

    :param function_name: the name of the function (use 'composite' if the function is a composite function)
    :param composite_function_expression: composite function specification such as
    ((((powerlaw{1} + (sin{2} * 3)) + (sin{2} * 25)) - (powerlaw{1} * 16)) + (sin{2} ** 3.0))
    :return: the an instance of the requested class

    """

    # Check whether this is a composite function or a simple function
    if composite_function_expression is not None:

        # Composite function

        return _parse_function_expression(composite_function_expression)

    else:

        if function_name in _known_functions:

            return _known_functions[function_name]()

        else:

            raise UnknownFunction("Function %s is not known. Known functions are: %s" %
                                  (function_name, ",".join(_known_functions.keys())))

def get_function_class(function_name):
    """
    Return the type for the requested function

    :param function_name: the function to return
    :return: the type for that function (i.e., this is a class, not an instance)
    """

    if function_name in _known_functions:

        return _known_functions[function_name]

    else:

        raise UnknownFunction("Function %s is not known. Known functions are: %s" %
                              (function_name, ",".join(_known_functions.keys())))


def list_functions():

    # Gather all defined functions and their descriptions

    functions_and_descriptions = {key:{'Description': value._function_definition['description']}
                                  for key,value in _known_functions.iteritems()}

    # Order by key (i.e., by function name)

    ordered = collections.OrderedDict(sorted(functions_and_descriptions.items()))

    # Format in a table

    table = dict_to_table(ordered)

    return table


def _parse_function_expression(function_specification):
    """
    Parse a complex function expression like:

    ((((powerlaw{1} + (sin{2} * 3)) + (sin{2} * 25)) - (powerlaw{1} * 16)) + (sin{2} ** 3.0))

    and return a composite function instance

    :param function_specification:
    :return: a composite function instance
    """

    # NOTE FOR SECURITY
    # This function has some security concerns. Security issues could arise if the user tries to read a model
    # file which has been maliciously formatted to contain harmful code. In this function we close all the doors
    # to a similar attack, except for those attacks which assume that the user has full access to a python environment.
    # Indeed, if that is the case, then the user can already do harm to the system, and so there is no point in
    # safeguard that from here. For example, the user could format a subclass of the Function class which perform
    # malicious operations in the constructor, add that to the dictionary of known functions, and then interpret
    # it with this code. However, if the user can instance malicious classes, then why would he use astromodels to
    # carry out the attack? Instead, what we explicitly check is the content of the function_specification string,
    # so that it cannot by itself do any harm (by for example containing instructions such as os.remove).

    # This can be a arbitrarily complex specification, like
    # ((((powerlaw{1} + (sin{2} * 3)) + (sin{2} * 25)) - (powerlaw{1} * 16)) + (sin{2} ** 3.0))

    # Use regular expressions to extract the set of functions like function_name{number},
    # then build the set of unique functions by using the constructor set()

    unique_functions = set(re.findall(r'\b([a-zA-Z0-9_]+)\{([0-9]?)\}',function_specification))

    # NB: unique functions is a set like:
    # {('powerlaw', '1'), ('sin', '2')}

    # Create instances of the unique functions

    instances = {}

    # Loop over the unique functions and create instances

    for (unique_function, number) in unique_functions:

        complete_function_specification = "%s{%s}" % (unique_function, number)

        # As first safety measure, check that the unique function is in the dictionary of _known_functions.
        # This could still be easily hacked, so it won't be the only check

        if unique_function in _known_functions:

            # Get the function class and check that it is indeed a proper Function class

            function_class = _known_functions[unique_function]

            if issubclass(function_class, Function):

                # Ok, let's create the instance

                instance = function_class()

                # Append the instance to the list

                instances[complete_function_specification] = instance

            else:

                raise FunctionDefinitionError("The function specification %s does not contain a proper function"
                                              % unique_function )

        else:

            raise UnknownFunction("Function %s in expression %s is unknown" % (unique_function, function_specification))

    # Check that we have found at least one instance.

    if len(instances)==0:

        raise DesignViolation("No known function in function specification")

    # The following presents a slight security problem if the model file that has been parsed comes from an untrusted
    # source. Indeed, the use of eval could make possible to execute things like os.remove.
    # In order to avoid this, first we substitute the function instances with numbers and remove the operators like
    # +,-,/ and so on. Then we try to execute the string with ast.literal_eval, which according to its documentation:

    # Safely evaluate an expression node or a Unicode or Latin-1 encoded string containing a Python literal or
    # container display. The string or node provided may only consist of the following Python literal structures:
    # strings, numbers, tuples, lists, dicts, booleans, and None.This can be used for safely evaluating strings
    # containing Python values from untrusted sources without the need to parse the values oneself.
    # It is not capable of evaluating arbitrarily complex expressions, for example involving operators or indexing.

    # If literal_eval cannot parse the string, it means that it contains unsafe input.

    # Create a copy of the function_specification

    string_for_literal_eval = function_specification

    # Remove from the function_specification all the known operators and function_expressions, and substitute them
    # with a 0 and a space

    # Let's start from the function expression

    for function_expression in instances.keys():

        string_for_literal_eval = string_for_literal_eval.replace(function_expression, '0 ')

    # Now remove all the known operators

    for operator in _operations.keys():

        string_for_literal_eval = string_for_literal_eval.replace(operator,'0 ')

    # The string at this point should contains only numbers and parenthesis separated by one or more spaces

    if re.match('''([a-zA-Z]+)''', string_for_literal_eval):

        raise DesignViolation("Extraneous input in function specification")

    # By using split() we separate all the numbers and parenthesis in a list, then we join them
    # with a comma, to end up with a comma-separated list of parenthesis and numbers like:
    # ((((0,0,(0,0,3)),0,(0,0,25)),0,(0,0,16)),0,(0,0,0,3.0))
    # This string can be parsed by literal_eval as a tuple containing other tuples, which is fine.
    # If the user has inserted some malicious content, like os.remove or more weird stuff like code objects,
    # the parsing will fail

    string_for_literal_eval = ",".join(string_for_literal_eval.split())

    #print(string_for_literal_eval)

    # At this point the string should be just a comma separated list of numbers

    # Now try to execute the string
    try:

        ast.literal_eval(string_for_literal_eval)

    except (ValueError, SyntaxError):

        raise DesignViolation("The given expression is not a valid function expression")

    else:

        # The expression is safe, let's eval it

        # First substitute the reference to the functions (like 'powerlaw{1}') with a string
        # corresponding to the instance dictionary

        sanitized_function_specification = function_specification

        for function_expression in instances.keys():

            sanitized_function_specification = sanitized_function_specification.replace(function_expression,
                                                                                        'instances["%s"]' %
                                                                                        function_expression)

        # Now eval it. For safety measure, I remove all globals, and the only local is the 'instances' dictionary

        composite_function = eval(sanitized_function_specification, {}, {'instances': instances})

        return composite_function


import numpy as np
from astropy.coordinates import SkyCoord, ICRS, BaseCoordinateFrame

from astromodels.functions.function import Function2D, FunctionMeta
from astromodels.utils.angular_distance import angular_distance


class Latitude_galactic_diffuse(Function2D):
    r"""
        description :

            A Gaussian distribution in Galactic latitude around the Galactic plane

        latex : $ K \exp{\left( \frac{-b^2}{2 \sigma_b^2} \right)} $

        parameters :

            K :

                desc : normalization
                initial value : 1

            sigma_b :

                desc : Sigma for
                initial value : 1

        """

    __metaclass__ = FunctionMeta

    # This is optional, and it is only needed if we need more setup after the
    # constructor provided by the meta class

    def _setup(self):

        self._frame = ICRS()

    def set_frame(self, new_frame):
        """
        Set a new frame for the coordinates (the default is ICRS J2000)

        :param new_frame: a coordinate frame from astropy
        :return: (none)
        """
        assert isinstance(new_frame, BaseCoordinateFrame)

        self._frame = new_frame

    def _set_units(self, x_unit, y_unit, z_unit):

        self.K.unit = z_unit
        self.sigma_b.unit = x_unit

    def evaluate(self, x, y, K, sigma_b):

        # We assume x and y are R.A. and Dec
        _coord = SkyCoord(ra=x, dec=y, frame=self._frame, unit="deg")

        b = _coord.transform_to('galactic').b.value

        return K * np.exp(-b ** 2 / (2 * sigma_b ** 2))


class Gaussian_on_sphere(Function2D):
    r"""
        description :

            A bidimensional Gaussian function on a sphere (in spherical coordinates)

        latex : $$ f(\vec{x}) = \left(\frac{180^\circ}{\pi}\right)^2 \frac{1}{2\pi \sqrt{\det{\Sigma}}} \, {\rm exp}\left( -\frac{1}{2} (\vec{x}-\vec{x}_0)^\intercal \cdot \Sigma^{-1}\cdot (\vec{x}-\vec{x}_0)\right) \\ \vec{x}_0 = ({\rm RA}_0,{\rm Dec}_0)\\ \Lambda = \left( \begin{array}{cc} \sigma^2 & 0 \\ 0 & \sigma^2 (1-e^2) \end{array}\right) \\ U = \left( \begin{array}{cc} \cos \theta & -\sin \theta \\ \sin \theta & cos \theta \end{array}\right) \\\Sigma = U\Lambda U^\intercal $$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            sigma :

                desc : Standard deviation of the Gaussian distribution
                initial value : 0.5
                min : 0
                max : 20

        """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit, z_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.sigma.unit = x_unit

    def evaluate(self, x, y, lon0, lat0, sigma):

        lon, lat = x,y

        angsep = angular_distance(lon0, lat0, lon, lat)

        return np.power(180 / np.pi, 2) * 1. / (2 * np.pi * sigma ** 2) * np.exp(
            -0.5 * np.power(angsep, 2) / sigma ** 2)

    def get_boundaries(self):

        # Truncate the gaussian at 2 times the max of sigma allowed

        max_sigma = self.sigma.max_value

        min_lat = max(-90., self.lat0.value - 2 * max_sigma)
        max_lat = min(90., self.lat0.value + 2 * max_sigma)

        max_abs_lat = max(np.absolute(min_lat), np.absolute(max_lat))

        if max_abs_lat > 89. or 2 * max_sigma / np.cos(max_abs_lat * np.pi / 180.) >= 180.:

            min_lon = 0.
            max_lon = 360.

        else:

            min_lon = self.lon0.value - 2 * max_sigma / np.cos(max_abs_lat * np.pi / 180.)
            max_lon = self.lon0.value + 2 * max_sigma / np.cos(max_abs_lat * np.pi / 180.)

            if min_lon < 0.:

                min_lon = min_lon + 360.

            elif max_lon > 360.:

                max_lon = max_lon - 360.

        return (min_lon, max_lon), (min_lat, max_lat)


class Disk_on_sphere(Function2D):
    r"""
        description :

            A bidimensional disk/tophat function on a sphere (in spherical coordinates)

        latex : $$ f(\vec{x}) = \left(\frac{180}{\pi}\right)^2 \frac{1}{\pi~({\rm radius})^2} ~\left\{\begin{matrix} 1 & {\rm if}& {\rm | \vec{x} - \vec{x}_0| \le {\rm radius}} \\ 0 & {\rm if}& {\rm | \vec{x} - \vec{x}_0| > {\rm radius}} \end{matrix}\right. $$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            radius :

                desc : Radius of the disk
                initial value : 0.5
                min : 0
                max : 20

        """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit, z_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.radius.unit = x_unit

    def evaluate(self, x, y, lon0, lat0, radius):

        lon, lat = x,y

        angsep = angular_distance(lon0, lat0, lon, lat)

        return np.power(180 / np.pi, 2) * 1. / (np.pi * radius ** 2) * (angsep <= radius)

    def get_boundaries(self):

        # Truncate the disk at 2 times the max of radius allowed

        max_radius = self.radius.max_value

        min_lat = max(-90., self.lat0.value - 2 * max_radius)
        max_lat = min(90., self.lat0.value + 2 * max_radius)

        max_abs_lat = max(np.absolute(min_lat), np.absolute(max_lat))

        if max_abs_lat > 89. or 2 * max_radius / np.cos(max_abs_lat * np.pi / 180.) >= 180.:

            min_lon = 0.
            max_lon = 360.

        else:

            min_lon = self.lon0.value - 2 * max_radius / np.cos(max_abs_lat * np.pi / 180.)
            max_lon = self.lon0.value + 2 * max_radius / np.cos(max_abs_lat * np.pi / 180.)

            if min_lon < 0.:

                min_lon = min_lon + 360.

            elif max_lon > 360.:

                max_lon = max_lon - 360.

        return (min_lon, max_lon), (min_lat, max_lat)


__author__ = 'giacomov'

from astromodels import sky_direction
from astromodels.functions import function
from astromodels import spectral_component
from astromodels import polarization
from astromodels.sources import point_source
from astromodels.sources import particle_source
from astromodels import parameter
from astromodels import model
from astromodels.my_yaml import my_yaml
from astromodels.sources.source import POINT_SOURCE, EXTENDED_SOURCE, PARTICLE_SOURCE
import re

class ModelIOError(IOError):
    pass


class ModelYAMLError(my_yaml.YAMLError):
    pass


class ModelSyntaxError(RuntimeError):
    pass


def load_model(filename):
    """
    Load a model from a file.

    :param filename: the name of the file containing the model
    :return: an instance of a Model
    """

    parser = ModelParser(filename)

    return parser.get_model()


class ModelParser(object):
    def __init__(self, model_file):

        # Read model file and deserialize into a dictionary

        try:

            with open(model_file) as f:

                self._model_dict = my_yaml.load(f)

        except IOError:

            raise ModelIOError("File %s cannot be read. Check path and permissions for current user." % model_file)

        except my_yaml.YAMLError:

            raise ModelYAMLError("Could not parse file %s. Check your syntax." % model_file)

        self._parse()

    def _parse(self):

        # Traverse the dictionary and create all the needed classes

        # The first level is the source level

        self._sources = []
        self._independent_variables = []
        self._links = []
        self._extra_setups = []

        for source_or_var_name, source_or_var_definition in self._model_dict.iteritems():

            if source_or_var_name.find("(IndependentVariable)") > 0:

                var_name = source_or_var_name.split("(")[0].replace(" ","")

                this_parser = IndependentVariableParser(var_name, source_or_var_definition)

                self._independent_variables.append(this_parser.get_variable())

            else:

                this_parser = SourceParser(source_or_var_name, source_or_var_definition)

                self._sources.append(this_parser.get_source())

                self._links.extend(this_parser.links)

                self._extra_setups.extend(this_parser.extra_setups)



    def get_model(self):

        # Instance the model with all the parsed sources

        new_model = model.Model(*self._sources)

        # Now set up IndependentVariable instances (if any)

        for independent_variable in self._independent_variables:

            new_model.add_independent_variable(independent_variable)

        # Now set up the links

        for link in self._links:

            path = link['parameter_path']
            variable = link['variable']
            law = link['law']

            new_model[path].add_auxiliary_variable(new_model[variable],law)

        # Finally the extra_setups (if any)

        for extra_setup in self._extra_setups:

            path = extra_setup['function_path']

            for property, value in extra_setup['extra_setup'].iteritems():

                new_model[path].__setattr__(property, new_model[value])

        return new_model


class IndependentVariableParser(object):

    def __init__(self, name, definition):

        self._variable = parameter.IndependentVariable(name, **definition)

    def get_variable(self):

        return self._variable


class SourceParser(object):
    def __init__(self, source_name, source_definition):

        # Get the type of the source

        try:

            # Point source or extended source?

            source_type = re.findall('\((%s|%s|%s)\)' % (POINT_SOURCE, EXTENDED_SOURCE, PARTICLE_SOURCE),
                                     source_name)[-1]

        except IndexError:

            raise ModelSyntaxError("Don't recognize type for source '%s'. "
                                   "Valid types are '%s', '%s' or '%s'." %
                                   (source_name, POINT_SOURCE, EXTENDED_SOURCE, PARTICLE_SOURCE))

        else:

            # Strip the source_type from the name

            source_name = source_name.split()[0]

        self._source_name = source_name

        # This will store the links (if any)
        self._links = []

        # This will store extra_setups (if any), used sometimes. For example, the function which uses naima
        # to make a synchrotron spectrum uses this to save and set up the particle distribution
        self._extra_setups = []

        if source_type == POINT_SOURCE:

            self._parsed_source = self._parse_point_source(source_definition)

        elif source_type == EXTENDED_SOURCE:

            self._parsed_source = self._parse_extended_source(source_definition)

        elif source_type == PARTICLE_SOURCE:

            self._parsed_source = self._parse_particle_source(source_definition)

    @property
    def extra_setups(self):

        return self._extra_setups

    @property
    def links(self):

        return self._links

    def get_source(self):

        return self._parsed_source

    def _parse_particle_source(self, particle_source_definition):

        # Parse the spectral information

        try:

            spectrum = particle_source_definition['spectrum']

        except KeyError:

            raise ModelSyntaxError("Point source %s is missing the 'spectrum' attribute" % self._source_name)

        components = []

        for component_name, component_definition in particle_source_definition['spectrum'].iteritems():

            this_component = self._parse_spectral_component(component_name, component_definition)

            components.append(this_component)

        this_particle_source = particle_source.ParticleSource(self._source_name, components=components)

        return this_particle_source

    def _parse_point_source(self, pts_source_definition):

        # Parse the positional information

        try:

            position_definition = pts_source_definition['position']

        except KeyError:

            raise ModelSyntaxError("Point source %s is missing the 'position' attribute" % self._source_name)

        this_sky_direction = self._parse_sky_direction(position_definition)

        # Parse the spectral information

        try:

            spectrum = pts_source_definition['spectrum']

        except KeyError:

            raise ModelSyntaxError("Point source %s is missing the 'spectrum' attribute" % self._source_name)

        components = []

        for component_name, component_definition in pts_source_definition['spectrum'].iteritems():
            this_component = self._parse_spectral_component(component_name, component_definition)

            components.append(this_component)

        this_point_source = point_source.PointSource(self._source_name, sky_position=this_sky_direction,
                                                     components=components)

        return this_point_source

    def _parse_sky_direction(self, sky_direction_definition):

        # Instance the SkyDirection class using the coordinates provided

        coordinates = {}

        if 'ra' in sky_direction_definition and 'dec' in sky_direction_definition:

            ra = parameter.Parameter('ra', sky_direction_definition['ra']['value'])
            ra.set_bounds(0, 360)
            ra.fix = True

            dec = parameter.Parameter('dec', sky_direction_definition['dec']['value'])
            dec.set_bounds(-90, 90)
            dec.fix = True

            coordinates['ra'] = ra
            coordinates['dec'] = dec

        elif 'l' in sky_direction_definition and 'b' in sky_direction_definition:

            l = parameter.Parameter('l', sky_direction_definition['l']['value'])
            l.set_bounds(0, 360)
            l.fix = True

            b = parameter.Parameter('b', sky_direction_definition['b']['value'])
            b.set_bounds(-90, 90)
            b.fix = True

            coordinates['l'] = l
            coordinates['b'] = b

        else:

            raise ModelSyntaxError("Position specification for source %s has an invalid coordinate pair. "
                                   " You need to specify either 'ra' and 'dec', or 'l' and 'b'."
                                   % self._source_name)

        # Check if there is a equinox specification

        if 'equinox' in sky_direction_definition:
            coordinates['equinox'] = sky_direction_definition['equinox']

        try:

            this_sky_direction = sky_direction.SkyDirection(**coordinates)

        except sky_direction.WrongCoordinatePair:

            raise ModelSyntaxError("Position specification for source %s has an invalid coordinate pair"
                                   % self._source_name)

        return this_sky_direction

    def _parse_spectral_component(self, component_name, component_definition):

        # Parse the shape definition, which is the first to occur

        try:

            function_name = component_definition.keys()[0]
            parameters_definition = component_definition[function_name]

        except KeyError:

            raise ModelSyntaxError("The component %s of source %s is malformed"
                                   % (component_name, self._source_name))

        shape = self._parse_shape_definition(component_name, function_name, parameters_definition)

        this_polarization = polarization.Polarization()

        this_spectral_component = spectral_component.SpectralComponent(component_name, shape, this_polarization)

        return this_spectral_component

    @staticmethod
    def _fix(value):
        # Remove new lines where it shouldn't be any
        # Sometimes YAML add new lines in the middle of definitions,
        # such as in units
        return value.replace("\n","")

    def _parse_shape_definition(self, component_name, function_name, parameters_definition):

        # Get the function

        if 'expression' in parameters_definition:

            # This is a composite function
            function_instance = function.get_function(function_name, parameters_definition['expression'])

        else:

            try:

                function_instance = function.get_function(function_name)

            except function.UnknownFunction:

                raise ModelSyntaxError("Function %s, specified as shape for %s of source %s, is not a "
                                       "known function" % (function_name, component_name, self._source_name))

        # Loop over the parameters of the function instance, instead of the specification,
        # so we can understand if there are parameters missing from the specification

        for parameter_name in function_instance.parameters.keys():

            try:

                this_definition = parameters_definition[parameter_name]

            except KeyError:

                raise ModelSyntaxError("Function %s, specified as shape for %s of source %s, lacks "
                                       "the definition for parameter %s"
                                       % (function_name, component_name, self._source_name, parameter_name))

            # Update the parameter. Note that the order is important, because trying to set the value before the
            # minimum and maximum could result in a error.

            # All these specifications are optional. If they are not present, then the default value
            # already contained in the instance of the function will be used

            if 'min' in this_definition:
                function_instance.parameters[parameter_name].min_value = this_definition['min']

            if 'max' in this_definition:
                function_instance.parameters[parameter_name].max_value = this_definition['max']

            if 'delta' in this_definition:
                function_instance.parameters[parameter_name].delta = this_definition['delta']

            if 'fix' in this_definition:
                function_instance.parameters[parameter_name].fix = this_definition['fix']

            if 'free' in this_definition:
                function_instance.parameters[parameter_name].free = this_definition['free']

            if 'unit' in this_definition:
                function_instance.parameters[parameter_name].unit = self._fix(this_definition['unit'])

            # Now set the value, which must be present

            if 'value' not in this_definition:

                raise ModelSyntaxError("The parameter %s in function %s, specified as shape for %s "
                                       "of source %s, lacks a 'value' attribute"
                                       % (parameter_name, function_name, component_name, self._source_name))

            # Check if this is a linked parameter, i.e., if 'value' is something like f(source.spectrum.powerlaw.index)

            matches = re.findall('''f\((.+)\)''', str(this_definition['value']))

            if matches:

                # This is an expression which marks a parameter
                # with a link to another parameter (or an IndependentVariable such as time)

                # Get the variable
                linked_variable = matches[0]

                # Now get the law

                if 'law' not in this_definition:

                    raise ModelSyntaxError("The parameter %s in function %s, specified as shape for %s "
                                           "of source %s, is linked to %s but lacks a 'law' attribute"
                                           % (parameter_name, function_name, component_name,
                                              self._source_name, linked_variable))

                link_function_name = this_definition['law'].keys()[0]

                link_function_instance = self._parse_shape_definition(component_name, link_function_name,
                                                                      this_definition['law'][link_function_name])

                path = ".".join([self._source_name, 'spectrum', component_name, function_name, parameter_name])

                self._links.append( {'parameter_path': path,
                                     'law': link_function_instance,
                                     'variable': linked_variable} )

            else:

                # This is a normal (not linked) parameter

                function_instance.parameters[parameter_name].value = this_definition['value']

            # Setup the prior for this parameter, if it exists
            if 'prior' in this_definition:

                # Get the function for this prior

                # A name to display in case of errors

                name_for_errors = 'prior for %s' % function_instance.parameters[parameter_name].path

                prior_function_name = this_definition['prior'].keys()[0]

                prior_function_definition = this_definition['prior'][prior_function_name]

                prior_function = self._parse_shape_definition(name_for_errors,
                                                              prior_function_name,
                                                              prior_function_definition)

                # Set it as prior for current parameter

                function_instance.parameters[parameter_name].prior = prior_function

        # Now handle extra_setup if any
        if 'extra_setup' in parameters_definition:

            path = ".".join([self._source_name, 'spectrum', component_name, function_name])

            self._extra_setups.append({'function_path': path,
                                       'extra_setup': parameters_definition['extra_setup']})


        return function_instance

    def _parse_extended_source(self, ext_source_definition):

        return 0



__author__ = 'giacomov'

from astropy.units import Quantity
import warnings


class ProtectedAttribute(RuntimeError):
    pass


class NonExistingAttribute(RuntimeWarning):
    pass


class DualAccessClass(object):
    """
    Suppose there is a class A having a dictionary b as a member. Inheriting from this class will allow the user to
    access the elements of b either as A.b[key] or as A.key. Accessing the element as A.key is read-only,
    i.e., A.key=something will fail.
    """

    def __init__(self, label, dictionary):

        self._lookup_dictionary = dictionary
        self._dictionary_label = label

        for key, value in self._lookup_dictionary.iteritems():

            super(DualAccessClass, self).__setattr__(key, value)

    def __setattr__(self, key, value):

        if hasattr(self, '_lookup_dictionary') and key in self._lookup_dictionary:

            item = self._lookup_dictionary[key]

            if hasattr(item, 'value') and not isinstance(item, Quantity):

                item.value = value

            else:

                raise ProtectedAttribute("You cannot assign to a %s" % self._dictionary_label)

        else:

            # Attributes which start with "_" are created by children classes

            if not key[0] == '_' and not hasattr(self, key):

                warnings.warn("Attribute %s does not exist. Check for typos." % key, NonExistingAttribute)

            super(DualAccessClass, self).__setattr__(key, value)

    def _add_attribute(self, name, value):

        self._lookup_dictionary[name] = value

        super(DualAccessClass, self).__setattr__(name, value)

    def _del_attribute(self, name):

        super(DualAccessClass, self).__delattr__(name)

        return self._lookup_dictionary.pop(name)



__author__ = 'giacomov'

# Import IPython display facility, if available. Otherwise,
# create a wrapper which just uses print

try:

    from IPython.display import display

except ImportError:

    def display(*args):
        """
        Mock version of display, used if there is no ipython installed
        """
        print(args)

try:

    from IPython.display import Latex

except ImportError:

    class Latex(object):
        """
        Mock version of the IPython Latex object, used if there is no ipython installed
        """

        def __init__(self, *args, **kwargs):

            pass

        def __repr__(self, *args, **kwargs):

            print("[you need to install IPython to see the Latex representation]")



#!/usr/bin/env python

import os
import sys
import glob
import numpy as np
import re

from setuptools import setup, Extension

# Get the version number
execfile('astromodels/version.py')


def setup_xspec():

    headas_root = os.environ.get("HEADAS")

    if headas_root is None:

        print("No HEADAS env. variable set. Xspec support will not be installed ")

        return None

    library_dirs = [os.path.join(headas_root, 'lib')]

    # Check that the library directories exist

    if not os.path.exists(library_dirs[0]):

        print("\nERROR: the library directory %s of HEADAS does not exist!" % library_dirs[0])

        sys.exit(-1)

    libraries = ['XSFunctions', 'XSModel', 'XSUtil', 'XS']

    # Check that the libraries exist
    for library in libraries:
        
        # Check for Linux/UNIX
        
        library_name = 'lib%s.so' % library

        if not os.path.exists(os.path.join(library_dirs[0], library_name)):

            # See if it has been compiled static (usually libXS is)

            library_name = 'lib%s.a' % library

            if not os.path.exists(os.path.join(library_dirs[0], library_name)):

                # If on OS X look for dylib
                
                library_name = 'lib%s.dylib' % library

                if not os.path.exists(os.path.join(library_dirs[0], library_name)):


                    print("\nERROR: the library %s does not exist in %s while setting up Xspec!" %
                          (library, library_dirs[0]))

    # Now find versions for required "external" libraries (part of HEASOFT but not Xspec by itself)

    required_libraries = ['cfitsio_', 'CCfits_', 'wcs-']
    

    for library_to_probe in required_libraries:
        
        # Linux/UNIX
        
        search_path = os.path.join(library_dirs[0], 'lib%s*.so' % library_to_probe)
        
        versions = glob.glob(search_path)

        if len(versions) == 0:
            
            search_path = os.path.join(library_dirs[0], 'lib%s*.so' % (library_to_probe[:-1]+'.'))
            
            versions = glob.glob(search_path)

            if len(versions) == 0:
                
                # Mac
                
                search_path = os.path.join(library_dirs[0], 'lib%s*.dylib' % library_to_probe)
                
                versions = glob.glob(search_path)

                if len(versions) == 0:

                    search_path = os.path.join(library_dirs[0], 'lib%s*.dylib' % (library_to_probe[:-1]+'.'))
                    
                    versions = glob.glob(search_path)

                    if len(versions) == 0:

                        print("\nERROR: cannot find version for library %s while setting up Xspec" % (library_to_probe))
                        sys.exit(-1)

        # Up to there versions[0] is a fully-qualified path
        # we need instead just the name of the library, without
        # the lib prefix nor the .so nor .a nor .dylib extension

        name = os.path.basename(versions[0])
        sanitized_name = re.match('lib(.+)\.', name).groups()[0]

        libraries.append(sanitized_name)

    # We also need gfortran, which must be installed on his own. If there is Xspec installed,
    # then there is also gfortran somewhere because it is a dependence
    libraries.append('gfortran')

    # Configure the variables to build the external module with the C/C++ wrapper

    ext_modules_configuration = [

        Extension("astromodels.xspec._xspec",

                  ["astromodels/xspec/src/_xspec.cc", ],

                  libraries=libraries,

                  include_dirs=['astromodels/xspec/include', np.get_include()],

                  library_dirs=library_dirs,
                  extra_compile_args=[])]

    return ext_modules_configuration

# Normal packages

packages = ['astromodels',
            'astromodels/functions',
            'astromodels/sources',
            'astromodels/utils',
            'astromodels/xspec'
            ]

# Check whether we can compile Xspec support
ext_modules_configuration = setup_xspec()

setup(
    name="astromodels",

    packages=packages,

    data_files=[('astromodels/data/functions', glob.glob('astromodels/data/functions/*.yaml'))],

    # The __version__ comes from the exec at the top

    version=__version__,

    description="Astromodels contains models to be used in likelihood or Bayesian analysis in astronomy",

    author='Giacomo Vianello',

    author_email='giacomo.vianello@gmail.com',

    url='https://github.com/giacomov/astromodels',

    download_url='https://github.com/giacomov/astromodels/archive/v0.1',

    keywords=['Likelihood', 'Models', 'fit'],

    classifiers=[],

    install_requires=[
        'numpy >= 1.6',
        'PyYAML',
        'astropy >= 1.0',
        'scipy>=0.13',
        'numdifftools'],

    ext_modules=ext_modules_configuration

)



__author__ = 'giacomov'

# The purpose of this module is to customize yaml so that it will load ordered dictionaries instead of normal
# ones. This way the order in which things are expressed in the file is maintained.

import yaml as my_yaml

import collections

_mapping_tag = my_yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG


def dict_representer(dumper, data):
    return dumper.represent_dict(data.iteritems())


def dict_constructor(loader, node):
    return collections.OrderedDict(loader.construct_pairs(node))


my_yaml.add_representer(collections.OrderedDict, dict_representer)
my_yaml.add_constructor(_mapping_tag, dict_constructor)



__author__ = 'giacomov'

import math
import numpy as np
import warnings
from scipy.special import gammaincc, gamma, erfcinv
import exceptions

from astromodels.functions.function import Function1D, Function2D, FunctionMeta, ModelAssertionViolation

from astromodels.units import get_units
import astropy.units as astropy_units


class GSLNotAvailable(UserWarning):
    pass


class NaimaNotAvailable(UserWarning):
    pass


class InvalidUsageForFunction(exceptions.Exception):
    pass

# Now let's try and import optional dependencies

try:

    # Naima is for numerical computation of Synch. and Inverse compton spectra in randomly oriented
    # magnetic fields

    import naima
    import astropy.units as u

except ImportError:

    warnings.warn("The naima package is not available. Models that depend on it will not be available",
                  NaimaNotAvailable)

    has_naima = False

else:

    has_naima = True

try:

    # GSL is the GNU Scientific Library. Pygsl is the python wrapper for it. It is used by some
    # functions for faster computation

    from pygsl.testing.sf import gamma_inc

except ImportError:

    warnings.warn("The GSL library or the pygsl wrapper cannot be loaded. Models that depend on it will not be "
                  "available.", GSLNotAvailable)

    has_gsl = False

else:

    has_gsl = True

# noinspection PyPep8Naming
class Powerlaw(Function1D):
    r"""
    description :

        A simple power-law

    latex : $ K~\frac{x}{piv}^{index} $

    parameters :

        K :

            desc : Normalization (differential flux at the pivot value)
            initial value : 1.0

        piv :

            desc : Pivot value
            initial value : 1
            fix : yes

        index :

            desc : Photon index
            initial value : -2
            min : -10
            max : 10

    tests :
        - { x : 10, function value: 0.01, tolerance: 1e-20}
        - { x : 100, function value: 0.0001, tolerance: 1e-20}

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # The index is always dimensionless
        self.index.unit = astropy_units.dimensionless_unscaled

        # The pivot energy has always the same dimension as the x variable
        self.piv.unit = x_unit

        # The normalization has the same units as the y

        self.K.unit = y_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, K, piv, index):

        xx = np.divide(x, piv)

        return K * np.power(xx, index)

# noinspection PyPep8Naming
class Powerlaw_flux(Function1D):
        r"""
        description :

            A simple power-law with the photon flux in a band used as normalization. This will reduce the correlation
            between the index and the normalization.

        latex : $ \frac{F(\gamma+1)} {b^{\gamma+1} - a^{\gamma+1}} (x)^{\gamma}$

        parameters :

            F :

                desc : Integral between a and b
                initial value : 1

            index :

                desc : Photon index
                initial value : -2
                min : -10
                max : 10

            a :

                desc : lower bound for the band in which computing the integral F
                initial value : 1.0
                fix : yes

            b :

                desc : upper bound for the band in which computing the integral F
                initial value : 100.0
                fix : yes

        """

        __metaclass__ = FunctionMeta

        def _set_units(self, x_unit, y_unit):

            # The flux is the integral over x, so:
            self.F.unit = y_unit * x_unit

            # The index is always dimensionless
            self.index.unit = astropy_units.dimensionless_unscaled

            # a and b have the same units as x

            self.a.unit = x_unit
            self.b.unit = x_unit

        # noinspection PyPep8Naming
        def evaluate(self, x, F, index, a, b):

            gp1 = index + 1

            return F * gp1 / (b**gp1 - a**gp1) * np.power(x, index)

class Cutoff_powerlaw(Function1D):
    r"""
    description :

        A power law multiplied by an exponential cutoff

    latex : $ K~\frac{x}{piv}^{index}~\exp{(-x/xc)} $

    parameters :

        K :

            desc : Normalization (differential flux at the pivot value)
            initial value : 1.0

        piv :

            desc : Pivot value
            initial value : 1
            fix : yes

        index :

            desc : Photon index
            initial value : -2
            min : -10
            max : 10

        xc :

            desc : Photon index
            initial value : 10.0
            min : 1.0

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # The index is always dimensionless
        self.index.unit = astropy_units.dimensionless_unscaled

        # The pivot energy has always the same dimension as the x variable
        self.piv.unit = x_unit

        # The cutoff has the same dimensions as x
        self.xc.unit = x_unit

        # The normalization has the same units as the y

        self.K.unit = y_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, K, piv, index, xc):

        return K * np.power(np.divide(x, piv), index) * np.exp(-1 * np.divide(x,xc))


class Broken_powerlaw(Function1D):
    r"""
    description :

        A broken power law function

    latex : $ f(x)= K~\begin{cases}\left( \frac{x}{x_{b}} \right)^{\alpha} & x < x_{b} \\ \left( \frac{x}{x_{b}} \right)^{\beta} & x \ge x_{b} \end{cases} $

    parameters :

        K :

            desc : Normalization (differential flux at x_b)
            initial value : 1.0

        xb :

            desc : Break point
            initial value : 10
            min : 1.0

        alpha :

            desc : Index before the break xb
            initial value : -1.5
            min : -10
            max : 10

        beta :

            desc : Index after the break xb
            initial value : -2.5
            min : -10
            max : 10

        piv :

            desc : Pivot energy
            initial value : 1.0
            fix : yes

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # The normalization has the same units as y
        self.K.unit = y_unit

        # The break point has always the same dimension as the x variable
        self.xb.unit = x_unit

        # alpha and beta are dimensionless
        self.alpha.unit = astropy_units.dimensionless_unscaled
        self.beta.unit = astropy_units.dimensionless_unscaled

        self.piv.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, K, xb, alpha, beta, piv):

        # The K * 0 is to keep the units right. If the input has unit, this will make a result
        # array with the same units as K. If the input has no units, this will have no
        # effect whatsoever

        result = np.zeros(x.shape) * K * 0

        idx = (x < xb)
        result[idx] = K * np.power(x[idx] / piv,alpha)
        result[~idx] = K * np.power(x[~idx] / piv,beta)

        return result

# noinspection PyPep8Naming
class Gaussian(Function1D):
    r"""
    description :

        A Gaussian function

    latex : $ K \frac{1}{\sigma \sqrt{2 \pi}}\exp{\frac{(x-\mu)^2}{2~(\sigma)^2}} $

    parameters :

        F :

            desc : Integral between -inf and +inf. Fix this to 1 to obtain a Normal distribution
            initial value : 1

        mu :

            desc : Central value
            initial value : 0.0

        sigma :

            desc : standard deviation
            initial value : 1.0
            min : 1e-12

    tests :
        - { x : 0.0, function value: 0.3989422804014327, tolerance: 1e-10}
        - { x : -1.0, function value: 0.24197072451914337, tolerance: 1e-9}

    """

    __metaclass__ = FunctionMeta

    # Place this here to avoid recomputing it all the time

    __norm_const = 1.0 / (math.sqrt(2 * np.pi))

    def _set_units(self, x_unit, y_unit):

        # The normalization is the integral from -inf to +inf, i.e., has dimensions of
        # y_unit * x_unit
        self.F.unit = y_unit * x_unit

        # The mu has the same dimensions as the x
        self.mu.unit = x_unit

        # sigma has the same dimensions as x
        self.sigma.unit = x_unit

    # noinspection PyPep8Naming
    def evaluate(self, x, F, mu, sigma):

        norm = self.__norm_const / sigma

        return F * norm * np.exp(-np.power(x - mu, 2.) / (2 * np.power(sigma, 2.)))

    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        mu = self.mu.value
        sigma = self.sigma.value

        sqrt_two = 1.414213562

        if x < 1e-16 or (1-x) < 1e-16:

            res = -1e32

        else:

            res = mu + sigma * sqrt_two * erfcinv(2 * (1-x))

        return res

class Uniform_prior(Function1D):
    r"""
    description :

        A function which is constant on the interval lower_bound - upper_bound and 0 outside the interval. The
        extremes of the interval are counted as part of the interval.

    latex : $ f(x)=\begin{cases}0 & x < \text{lower_bound} \\\text{value} & \text{lower_bound} \le x \le \text{upper_bound} \\ 0 & x > \text{upper_bound} \end{cases}$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : 0
            min : -np.inf
            max : np.inf

        upper_bound :

            desc : Upper bound for the interval
            initial value : 1
            min : -np.inf
            max : np.inf

        value :

            desc : Value in the interval
            initial value : 1.0

    tests :
        - { x : 0.5, function value: 1.0, tolerance: 1e-20}
        - { x : -0.5, function value: 0, tolerance: 1e-20}

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # Lower and upper bound has the same unit as x
        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit

        # value has the same unit as y
        self.value.unit = y_unit

    def evaluate(self, x, lower_bound, upper_bound, value):

        # The value * 0 is to keep the units right

        result = np.zeros(x.shape) * value * 0

        idx = (x >= lower_bound) & (x <= upper_bound)
        result[idx] = value

        return result

    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        lower_bound = self.lower_bound.value
        upper_bound = self.upper_bound.value

        low = lower_bound
        spread = float(upper_bound - lower_bound)

        par = x * spread + low

        return par


class Log_uniform_prior(Function1D):
    r"""
    description :

        A function which is 1/x on the interval lower_bound - upper_bound and 0 outside the interval. The
        extremes of the interval are NOT counted as part of the interval. Lower_bound must be >= 0.

    latex : $ f(x)=\begin{cases}0 & x \le \text{lower_bound} \\\frac{1}{x} & \text{lower_bound} < x < \text{upper_bound} \\ 0 & x \ge \text{upper_bound} \end{cases}$

    parameters :

        lower_bound :

            desc : Lower bound for the interval
            initial value : 1e-20
            min : 1e-30
            max : np.inf

        upper_bound :

            desc : Upper bound for the interval
            initial value : 100
            min : 1e-30
            max : np.inf

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # Lower and upper bound has the same unit as x
        self.lower_bound.unit = x_unit
        self.upper_bound.unit = x_unit

    def evaluate(self, x, lower_bound, upper_bound):

        # This makes the prior proper because it is the integral between lower_bound and upper_bound
        # Assume we are in the "fast" mode, where things have no units, and fall back to the slow mode
        # if this isn't true
        #try:

        #    renorm = math.log(upper_bound) - math.log(lower_bound)

        #except (astropy_units.UnitsError, TypeError):

        #    renorm = math.log(upper_bound.value) - math.log(lower_bound.value)

        renorm = 1

        result = np.zeros_liks(x)

        idx = (x > lower_bound) & (x < upper_bound)
        result[idx] = 1.0/x
        result[~idx] = 0.0

        return 1.0 / renorm * result

    def from_unit_cube(self, x):
        """
        Used by multinest

        :param x: 0 < x < 1
        :param lower_bound:
        :param upper_bound:
        :return:
        """

        low = math.log10(self.lower_bound.value)
        up = math.log10(self.upper_bound.value)

        spread = up - low
        par = 10**(x * spread + low)

        return par


# noinspection PyPep8Naming
class Sin(Function1D):
    r"""
    description :

        A sinusodial function

    latex : $ K~\sin{(2\pi f x + \phi)} $

    parameters :

        K :

            desc : Normalization
            initial value : 1

        f :

            desc : frequency
            initial value : 1.0 / (2 * np.pi)
            min : 0

        phi :

            desc : phase
            initial value : 0
            min : -np.pi
            max : +np.pi
            unit: rad

    tests :
        - { x : 0.0, function value: 0.0, tolerance: 1e-10}
        - { x : 1.5707963267948966, function value: 1.0, tolerance: 1e-10}

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # The normalization has the same unit of y
        self.K.unit = y_unit

        # The unit of f is 1 / [x] because fx must be a pure number. However,
        # np.pi of course doesn't have units, so we add a rad
        self.f.unit = x_unit**(-1) * astropy_units.rad

        # The unit of phi is always the same (radians)

        self.phi.unit = astropy_units.rad

    # noinspection PyPep8Naming
    def evaluate(self, x, K, f, phi):

        return K * np.sin(2 * np.pi * f * x + phi)


if has_naima:

    class Synchrotron(Function1D):
        r"""
        description :

            Synchrotron spectrum from an input particle distribution, using Naima (naima.readthedocs.org)

        latex: not available

        parameters :

            B :

                desc : magnetic field
                initial value : 3.24e-6
                unit: Gauss

            distance :

                desc : distance of the source
                initial value : 1.0
                unit : kpc

            emin :

                desc : minimum energy for the particle distribution
                initial value : 1
                fix : yes
                unit: GeV

            emax :
                desc : maximum energy for the particle distribution
                initial value : 510e3
                fix : yes
                unit: GeV

            need:

                desc: number of points per decade in which to evaluate the function
                initial value : 10
                min : 2
                max : 100
                fix : yes

        """

        __metaclass__ = FunctionMeta

        def _set_units(self, x_unit, y_unit):

            # This function can only be used as a spectrum,
            # so let's check that x_unit is a energy and y_unit is
            # differential flux

            if hasattr(x_unit,"physical_type") and x_unit.physical_type == 'energy':

                # Now check that y is a differential flux
                current_units = get_units()
                should_be_unitless = y_unit * (current_units.energy * current_units.time * current_units.area)

                if not hasattr(should_be_unitless,'physical_type') or \
                   should_be_unitless.decompose().physical_type != 'dimensionless':

                    # y is not a differential flux
                    raise InvalidUsageForFunction("Unit for y is not differential flux. The function synchrotron "
                                                  "can only be used as a spectrum.")
            else:

                raise InvalidUsageForFunction("Unit for x is not an energy. The function synchrotron can only be used "
                                              "as a spectrum")

            # we actually don't need to do anything as the units are already set up

        def set_particle_distribution(self, function):

            self._particle_distribution = function

            # Now set the units for the function

            current_units = get_units()

            self._particle_distribution.set_units(current_units.energy, current_units.energy**(-1))

            # Naima wants a function which accepts a quantity as x (in units of eV) and returns an astropy quantity,
            # so we need to create a wrapper which will remove the unit from x and add the unit to the return
            # value

            self._particle_distribution_wrapper = lambda x: function(x.value) / current_units.energy

        def get_particle_distribution(self):

            return self._particle_distribution

        particle_distribution = property(get_particle_distribution, set_particle_distribution,
                                         doc="""Get/set particle distribution for electrons""")

        # noinspection PyPep8Naming
        def evaluate(self, x, B, distance, emin, emax, need):

            _synch = naima.models.Synchrotron(self._particle_distribution_wrapper, B * astropy_units.Gauss,
                                              Eemin = emin * astropy_units.GeV,
                                              Eemax = emax * astropy_units.GeV, nEed = need)

            return _synch.flux(x * get_units().energy, distance=distance * astropy_units.kpc).value

        def to_dict(self, minimal=False):

            data = super(Function1D, self).to_dict(minimal)

            if not minimal:

                data['extra_setup'] = {'particle_distribution': self.particle_distribution.path}

            return data


class Line(Function1D):
    r"""
    description :

        A linear function

    latex : $ a * x + b $

    parameters :

        a :

            desc : linear coefficient
            initial value : 1

        b :

            desc : intercept
            initial value : 0

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # a has units of y_unit / x_unit, so that a*x has units of y_unit
        self.a.unit = y_unit / x_unit

        # b has units of y
        self.b.unit = y_unit

    def evaluate(self, x, a, b):

        return a * x + b


class Constant(Function1D):
    r"""
        description :

            Return k

        latex : $ k $

        parameters :

            k :

                desc : Constant value
                initial value : 0

        """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        self.k.unit = y_unit


    def evaluate(self, x, k):

        return k


class Bias(Function1D):
    r"""
    description :

        Return x plus a bias

    latex : $ x + k$

    parameters :

        k :

            desc : Constant value
            initial value : 0

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # k has units of x

        self.k.unit = x_unit

        if x_unit != y_unit:

            raise InvalidUsageForFunction("Function bias cannot be given different units for x and y")

    def evaluate(self, x, k):

        return x + k

class Band(Function1D):
    r"""
    description :

        Band model from Band et al., 1993, parametrized with the peak energy

    latex : $  $

    parameters :

        K :

            desc : Differential flux at the pivot energy
            initial value : 1e-4

        alpha :

            desc : low-energy photon index
            initial value : -1.0
            min : -1.5
            max : 3

        xp :

            desc : peak in the x * x * N (nuFnu if x is a energy)
            initial value : 500
            min : 10

        beta :

            desc : high-energy photon index
            initial value : -2.0
            min : -5.0
            max : -1.6

        piv :

            desc : pivot energy
            initial value : 100.0
            fix : yes
    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # The normalization has the same units as y
        self.K.unit = y_unit

        # The break point has always the same dimension as the x variable
        self.xp.unit = x_unit

        self.piv.unit = x_unit

        # alpha and beta are dimensionless
        self.alpha.unit = astropy_units.dimensionless_unscaled
        self.beta.unit = astropy_units.dimensionless_unscaled

    def evaluate(self, x, K, alpha, xp, beta, piv):

        E0 = xp / (2 + alpha)

        if (alpha < beta):
            raise ModelAssertionViolation("Alpha cannot be less than beta")

        idx = x < (alpha - beta) * E0

        # The K * 0 part is a trick so that out will have the right units (if the input
        # has units)

        out = np.zeros(x.shape) * K * 0

        out[idx] = K * np.power(x[idx] / piv, alpha) * np.exp(-x[idx] / E0)
        out[~idx] = K * np.power((alpha - beta) * E0 / piv, alpha - beta) * np.exp(beta - alpha) * \
                    np.power(x[~idx] / piv, beta)

        return out


class Band_Calderone(Function1D):
    r"""
    description :

        The Band model from Band et al. 1993, implemented however in a way which reduces the covariances between
        the parameters (Calderone et al., MNRAS, 448, 403C, 2015)

    latex : $ \text{(Calderone et al., MNRAS, 448, 403C, 2015)} $

    parameters :

        alpha :
            desc : The index for x smaller than the x peak
            initial value : -1
            min : -10
            max : 10

        beta :

            desc : index for x greater than the x peak (only if opt=1, i.e., for the
                   Band model)
            initial value : -2.2
            min : -7
            max : -1

        xp :

            desc : position of the peak in the x*x*f(x) space (if x is energy, this is the nuFnu or SED space)
            initial value : 200.0
            min : 0

        F :

            desc : integral in the band defined by a and b
            initial value : 1e-6

        a:

            desc : lower limit of the band in which the integral will be computed
            initial value : 1.0
            min : 0
            fix : yes

        b:

            desc : upper limit of the band in which the integral will be computed
            initial value : 10000.0
            min : 0
            fix : yes

        opt :

            desc : option to select the spectral model (0 corresponds to a cutoff power law, 1 to the Band model)
            initial value : 1
            min : 0
            max : 1
            fix : yes

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # alpha and beta are always unitless

        self.alpha.unit = astropy_units.dimensionless_unscaled
        self.beta.unit = astropy_units.dimensionless_unscaled

        # xp has the same dimension as x
        self.xp.unit = x_unit

        # F is the integral over x, so it has dimensions y_unit * x_unit
        self.F.unit = y_unit * x_unit

        # a and b have the same units of x
        self.a.unit = x_unit
        self.b.unit = x_unit

        # opt is just a flag, and has no units
        self.opt.unit = astropy_units.dimensionless_unscaled

    @staticmethod
    def ggrb_int_cpl( a, Ec, Emin, Emax):

        # Gammaincc does not support quantities
        i1 = gammaincc(2 + a, Emin / Ec) * gamma(2+a)
        i2 = gammaincc(2 + a, Emax / Ec) * gamma(2+a)

        return -Ec * Ec * (i2 - i1)

    @staticmethod
    def ggrb_int_pl(a, b, Ec, Emin, Emax):

        pre = pow(a-b, a-b) * math.exp(b-a) / pow(Ec, b)

        if b != -2:

            return pre / (2+b) * (pow(Emax, 2+b) - pow(Emin, 2+b))

        else:

            return pre * math.log(Emax / Emin)

    def evaluate(self, x, alpha, beta, xp, F, a, b, opt):

        assert opt == 0 or opt == 1, "Opt must be either 0 or 1"

        if alpha < beta:

            raise ModelAssertionViolation("Alpha cannot be smaller than beta")

        if alpha < -2:

            raise ModelAssertionViolation("Alpha cannot be smaller than -2")

        # Cutoff energy

        if alpha == -2:

            Ec = xp / 0.0001 #TRICK: avoid a=-2

        else:

            Ec = xp / (2 + alpha)

        # Split energy

        Esplit = (alpha-beta) * Ec

        # Evaluate model integrated flux and normalization

        if isinstance(alpha, astropy_units.Quantity):

            # The following functions do not allow the use of units
            alpha_ = alpha.value
            Ec_ = Ec.value
            a_ = a.value
            b_ = b.value
            Esplit_ = Esplit.value
            beta_ = beta.value

            unit_ = self.x_unit

        else:

            alpha_, Ec_, a_, b_, Esplit_, beta_ = alpha, Ec, a, b, Esplit, beta
            unit_ = 1.0

        if opt==0:

            # Cutoff power law

            intflux = self.ggrb_int_cpl(alpha_, Ec_, a_, b_)

        else:

            # Band model

            if a <= Esplit and Esplit <= b:

                intflux = (self.ggrb_int_cpl(alpha_, Ec_, a_, Esplit_) +
                           self.ggrb_int_pl (alpha_, beta_, Ec_, Esplit_, b_))

            else:

                if Esplit < a:

                    intflux = self.ggrb_int_pl(alpha_, beta_, Ec_, a_, b_)

                else:

                    raise RuntimeError("Esplit > emax!")

        erg2keV = 6.24151e8

        norm = F * erg2keV / (intflux * unit_)

        if opt==0:

            # Cutoff power law

            flux = np.power(x / Ec, alpha) * np.exp( - x / Ec)

        else:

            # The norm * 0 is to keep the units right

            flux = np.zeros(x.shape) * norm * 0

            idx = x < Esplit

            flux[idx] = norm *  np.power(x[idx] / Ec, alpha) * np.exp(-x[idx] / Ec)
            flux[~idx] = norm * pow(alpha - beta, alpha - beta) * math.exp(beta - alpha) * np.power(x[~idx] / Ec, beta)

        return flux


class Log_parabola(Function1D):
    r"""
    description :

        A log-parabolic function

    latex : $ K \left( \frac{x}{piv} \right)^{\alpha +\beta \log{\left( \frac{x}{piv} \right)}} $

    parameters :

        K :

            desc : Normalization
            initial value : 1.0

        piv :
            desc : Pivot (keep this fixed)
            initial value : 1
            fix : yes

        alpha :

            desc : index
            initial value : -2.0

        beta :

            desc : curvature (negative is concave, positive is convex)
            initial value : -1.0

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):

        # K has units of y

        self.K.unit = y_unit

        # piv has the same dimension as x
        self.piv.unit = x_unit

        # alpha and beta are dimensionless
        self.alpha.unit = astropy_units.dimensionless_unscaled
        self.beta.unit = astropy_units.dimensionless_unscaled

    def evaluate(self, x, K, piv, alpha, beta):

        xx = np.divide(x, piv)

        try:

            return K * xx**(alpha + beta * np.log10(xx))

        except ValueError:

            # The current version of astropy (1.1.x) has a bug for which quantities that have become
            # dimensionless because of a division (like xx here) are not recognized as such by the power
            # operator, which throws an exception: ValueError: Quantities and Units may only be raised to a scalar power
            # This is a quick fix, waiting for astropy 1.2 which will fix this

            xx = xx.to('')

            return K * xx**(alpha + beta * np.log10(xx))

    @property
    def peak_energy(self):
        """
        Returns the peak energy in the nuFnu spectrum

        :return: peak energy in keV
        """

        # Eq. 6 in Massaro et al. 2004
        # (http://adsabs.harvard.edu/abs/2004A%26A...413..489M)

        return self.piv.value * pow(10, (2 + self.alpha.value) / (2 * self.beta.value) )

if has_gsl:

    class Cutoff_powerlaw_flux(Function1D):
        r"""
            description :

                A cutoff power law having the flux as normalization, which should reduce the correlation among
                parameters.

            latex : $ \frac{F}{T(b)-T(a)} ~x^{index}~\exp{(-x/x_{c})}~\text{with}~T(x)=-x_{c}^{index+1} \Gamma(index+1, x/C)~\text{(}\Gamma\text{ is the incomplete gamma function)} $

            parameters :

                F :

                    desc : Integral between a and b
                    initial value : 1e-5

                index :

                    desc : photon index
                    initial value : -2.0

                xc :

                    desc : cutoff position
                    initial value : 50.0

                a :

                    desc : lower bound for the band in which computing the integral F
                    initial value : 1.0
                    fix : yes

                b :

                    desc : upper bound for the band in which computing the integral F
                    initial value : 100.0
                    fix : yes
            """

        __metaclass__ = FunctionMeta

        def _set_units(self, x_unit, y_unit):

            # K has units of y * x
            self.F.unit = y_unit * x_unit

            # alpha is dimensionless
            self.index.unit = astropy_units.dimensionless_unscaled

            # xc, a and b have the same dimension as x
            self.xc.unit = x_unit
            self.a.unit = x_unit
            self.b.unit = x_unit

        @staticmethod
        def _integral(a, b, index, ec):

            ap1 = index + 1

            integrand = lambda x: -pow(ec, ap1) * gamma_inc(ap1, x / ec)

            return integrand(b) - integrand(a)

        def evaluate(self, x, F, index, xc, a, b):

            this_integral = self._integral(a, b, index, xc)

            return F / this_integral * np.power(x, index) * np.exp(-1 * np.divide(x, xc))


class Exponential_cutoff(Function1D):
    r"""
        description :

            An exponential cutoff

        latex : $ K \exp{(-x/xc)} $

        parameters :

            K :

                desc : Normalization
                initial value : 1.0
                fix : yes

            xc :
                desc : cutoff
                initial value : 100
                min : 1
        """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit):
        # K has units of y

        self.K.unit = y_unit

        # piv has the same dimension as x
        self.xc.unit = x_unit

    def evaluate(self, x, K, xc):

        return K * np.exp(np.divide(x, -xc))
